from xdsl.ir import Dialect
from xdsl.passes import ModulePass
from xdsl.universe import Universe


class PluginPass(ModulePass):
    name = "plugin-pass"


PluginDialect = Dialect("plugin_dialect")

MY_UNIVERSE = Universe(
    all_dialects={"plugin_dialect": lambda: PluginDialect},
    all_passes={"plugin-pass": lambda: PluginPass},
)



import math
from collections.abc import Sequence
from typing import Any, cast

from typing_extensions import TypeVar

from xdsl.dialects import tensor
from xdsl.dialects.builtin import DYNAMIC_INDEX, TensorType
from xdsl.interpreter import (
    Interpreter,
    InterpreterFunctions,
    impl,
    register_impls,
)
from xdsl.interpreters.builtin import xtype_for_el_type
from xdsl.interpreters.shaped_array import ShapedArray
from xdsl.interpreters.utils.ptr import TypedPtr
from xdsl.utils.exceptions import InterpretationError

T = TypeVar("T")


@register_impls
class TensorFunctions(InterpreterFunctions):
    @impl(tensor.EmptyOp)
    def run_empty(
        self, interpreter: Interpreter, op: tensor.EmptyOp, args: tuple[int, ...]
    ) -> tuple[ShapedArray[Any]]:
        result_type = op.tensor.type
        assert isinstance(result_type, TensorType)
        result_shape = list(result_type.get_shape())

        dynamic_dims = iter(args)
        for i in range(len(result_shape)):
            if result_shape[i] == DYNAMIC_INDEX:
                result_shape[i] = next(dynamic_dims)
        assert next(dynamic_dims, None) is None

        xtype = xtype_for_el_type(result_type.element_type, interpreter.index_bitwidth)
        return (
            ShapedArray(
                TypedPtr[Any].new((0,) * math.prod(result_shape), xtype=xtype),
                result_shape,
            ),
        )

    @impl(tensor.ReshapeOp)
    def run_reshape(
        self,
        interpreter: Interpreter,
        op: tensor.ReshapeOp,
        args: tuple[ShapedArray[T], ShapedArray[int]],
    ) -> tuple[ShapedArray[T]]:
        input, new_shape = args
        assert isinstance(input, ShapedArray)
        result_type = op.result.type
        assert isinstance(result_type, TensorType)
        static_shape = list(result_type.get_shape())
        assert static_shape is not None
        if static_shape != new_shape.data:
            raise InterpretationError("Mismatch between static shape and new shape")
        result = ShapedArray(input.data_ptr, static_shape)
        return (result,)

    @impl(tensor.InsertOp)
    def run_insert(
        self,
        interpreter: Interpreter,
        op: tensor.InsertOp,
        args: tuple[T | ShapedArray[T] | int, ...],
    ) -> tuple[ShapedArray[T]]:
        value = cast(T, args[0])
        dest = cast(ShapedArray[T], args[1])
        indices = cast(Sequence[int], args[2:])

        assert isinstance(dest, ShapedArray)
        assert len(indices) == len(dest.shape)

        result = dest.copy()
        result.store(indices, value)

        return (result,)

    @impl(tensor.ExtractOp)
    def run_extract(
        self,
        interpreter: Interpreter,
        op: tensor.ExtractOp,
        args: tuple[ShapedArray[T] | int, ...],
    ) -> tuple[T]:
        tensor = cast(ShapedArray[T], args[0])
        indices = cast(Sequence[int], args[1:])

        assert isinstance(tensor, ShapedArray)
        assert len(indices) == len(tensor.shape)

        return (tensor.load(indices),)

    @impl(tensor.DimOp)
    def run_dim(
        self,
        interpreter: Interpreter,
        op: tensor.DimOp,
        args: tuple[ShapedArray[T] | int, ...],
    ) -> tuple[int]:
        tensor = cast(ShapedArray[T], args[0])
        dim = cast(int, args[1])

        assert isinstance(tensor, ShapedArray)

        return (tensor.shape[dim],)



import pytest

from xdsl.dialects.arm_neon import (
    NeonArrangement,
    NeonArrangementAttr,
    NEONRegisterType,
    vector_with_arrangement,
)
from xdsl.dialects.builtin import VectorType, f16, f32, f64


def test_assembly_str_without_index():
    reg = NEONRegisterType.from_name("v0")
    arrangement = NeonArrangementAttr(NeonArrangement.D)
    vecreg = vector_with_arrangement(reg=reg, arrangement=arrangement, index=None)
    assert vecreg == "v0.2D"


def test_assembly_str_with_index():
    reg = NEONRegisterType.from_name("v0")
    arrangement = NeonArrangementAttr(NeonArrangement.D)
    vecreg = vector_with_arrangement(reg=reg, arrangement=arrangement, index=5)
    assert vecreg == "v0.D[5]"


def test_arr_from_vec():
    assert NeonArrangement.from_vec_type(VectorType(f16, [8])) == NeonArrangement.H
    assert NeonArrangement.from_vec_type(VectorType(f32, [4])) == NeonArrangement.S
    assert NeonArrangement.from_vec_type(VectorType(f64, [2])) == NeonArrangement.D

    with pytest.raises(
        ValueError, match="Invalid vector type for ARM NEON: vector<3xf16>"
    ):
        NeonArrangement.from_vec_type(VectorType(f16, [3]))



import re
from collections.abc import Callable, Sequence

import pytest
from typing_extensions import TypeVar

from xdsl.ir.affine import (
    AffineBinaryOpExpr,
    AffineBinaryOpKind,
    AffineExpr,
    AffineMap,
    SimpleAffineExprFlattener,
)


def test_simple_map():
    # x, y
    x = AffineExpr.dimension(0)
    y = AffineExpr.dimension(1)

    # map1: (x, y) -> (x + y, y)
    map1 = AffineMap(2, 0, (x + y, y))
    assert map1.eval([1, 2], []) == (3, 2)
    assert map1.eval([3, 4], []) == (7, 4)
    assert map1.eval([5, 6], []) == (11, 6)

    # map2: (x, y) -> (2x + 3y)
    map2 = AffineMap(2, 0, (2 * x + 3 * y,))
    assert map2.eval([1, 2], []) == (8,)
    assert map2.eval([3, 4], []) == (18,)
    assert map2.eval([5, 6], []) == (28,)

    # map3: (x, y) -> (x + y, 2x + 3y)
    map3 = AffineMap(2, 0, (x + y, 2 * x + 3 * y))
    assert map3.eval([1, 2], []) == (3, 8)
    assert map3.eval([3, 4], []) == (7, 18)
    assert map3.eval([5, 6], []) == (11, 28)


def test_quasiaffine_map():
    # x
    x = AffineExpr.dimension(0)
    # N
    N = AffineExpr.symbol(0)

    # map1: (x)[N] -> (x floordiv 2)
    map1 = AffineMap(1, 1, (x // 2,))
    assert map1.eval([1], [10]) == (0,)
    assert map1.eval([2], [10]) == (1,)
    assert map1.eval([3], [10]) == (1,)
    assert map1.eval([4], [13]) == (2,)
    assert map1.eval([5], [10]) == (2,)
    assert map1.eval([6], [11]) == (3,)

    # map2: (x)[N] -> (-(x ceildiv 2) + N)
    map2 = AffineMap(1, 1, (-(x.ceil_div(2)) + N,))
    assert map2.eval([1], [10]) == (9,)
    assert map2.eval([2], [10]) == (9,)
    assert map2.eval([3], [10]) == (8,)
    assert map2.eval([4], [13]) == (11,)
    assert map2.eval([5], [10]) == (7,)
    assert map2.eval([6], [11]) == (8,)

    # map3: (x)[N] -> (x mod 2 - N)
    map3 = AffineMap(1, 1, ((x % 2) - N,))
    assert map3.eval([1], [10]) == (-9,)
    assert map3.eval([2], [10]) == (-10,)
    assert map3.eval([3], [10]) == (-9,)
    assert map3.eval([4], [13]) == (-13,)
    assert map3.eval([5], [10]) == (-9,)
    assert map3.eval([6], [11]) == (-11,)


def test_composition_simple():
    # map1 = (x, y) -> (x - y)
    map1 = AffineMap(2, 0, (AffineExpr.dimension(0) - AffineExpr.dimension(1),))
    # map2 = (x, y) -> (y, x)
    map2 = AffineMap(
        2,
        0,
        (AffineExpr.dimension(1), AffineExpr.dimension(0)),
    )
    # Compose
    # map3 = (x, y) -> (y - x)
    map3 = map1.compose(map2)

    assert map3.eval([1, 2], []) == (1,)
    assert map3.eval([3, 4], []) == (1,)
    assert map3.eval([5, 6], []) == (1,)
    assert map3.eval([20, 10], []) == (-10,)


def test_composition():
    # map1: (x, y) -> (x floordiv 2, 2 * x + 3 * y)
    map1 = AffineMap(
        2,
        0,
        (
            AffineExpr.dimension(0) // 2,
            2 * AffineExpr.dimension(0) + 3 * AffineExpr.dimension(1),
        ),
    )
    # map2: (x, y) -> (-x, -y)
    map2 = AffineMap(2, 0, (-AffineExpr.dimension(0), -AffineExpr.dimension(1)))
    # Compose
    # map3: (x, y) -> (-x floordiv 2, -2 * x - 3 * y)
    map3 = map1.compose(map2)

    assert map3.eval([1, 2], []) == (-1, -8)
    assert map3.eval([3, 4], []) == (-2, -18)
    assert map3.eval([5, 6], []) == (-3, -28)


def test_compose_expr():
    d = [AffineExpr.dimension(i) for i in range(3)]
    s = [AffineExpr.symbol(i) for i in range(2)]

    expr = d[0] + d[2]
    map = AffineMap.from_callable(
        lambda d0, d1, d2, s0, s1: (d0 + s1, d1 + s0, d0 + d1 + d2),
        dim_symbol_split=(3, 2),
    )
    expected = (d[0] + s[1]) + (d[0] + d[1] + d[2])
    assert expr.compose(map) == expected


def test_compose_expr_recursive_simplification():
    d = [AffineExpr.dimension(i) for i in range(4)]

    # Tests simplifications that require recursion
    add1 = (d[0] + d[1]) + (d[2] + d[3])
    assert add1.compose(
        AffineMap.from_callable(lambda d0, d1: (1, 2, 3, 4))
    ) == AffineExpr.constant(10)

    add2 = d[0] + d[1] + d[2] + d[3]
    assert add2.compose(
        AffineMap.from_callable(lambda d0, d1: (1, 2, 3, 4))
    ) == AffineExpr.constant(10)


def test_compose_map():
    map1 = AffineMap.from_callable(
        lambda d0, d1, s0, s1: (d0 + 1 + s1, d1 - 1 - s0), dim_symbol_split=(2, 2)
    )
    map2 = AffineMap.from_callable(
        lambda d0, s0: (d0 + s0, d0 - s0), dim_symbol_split=(1, 1)
    )
    map3 = AffineMap.from_callable(
        lambda d0, s0, s1, s2: ((d0 + s2) + 1 + s1, (d0 - s2) - 1 - s0),
        dim_symbol_split=(1, 3),
    )

    assert map1.compose(map2) == map3


def test_helpers():
    m = AffineMap.constant_map(0)
    assert m == AffineMap(0, 0, (AffineExpr.constant(0),))
    m = AffineMap.point_map(0, 1)
    assert m == AffineMap(0, 0, (AffineExpr.constant(0), AffineExpr.constant(1)))
    m = AffineMap.identity(2)
    assert m == AffineMap(2, 0, (AffineExpr.dimension(0), AffineExpr.dimension(1)))
    m = AffineMap.identity(0, 2)
    assert m == AffineMap(0, 2, (AffineExpr.symbol(0), AffineExpr.symbol(1)))
    m = AffineMap.identity(2, 2)
    assert m == AffineMap(
        2,
        2,
        (
            AffineExpr.dimension(0),
            AffineExpr.dimension(1),
            AffineExpr.symbol(0),
            AffineExpr.symbol(1),
        ),
    )
    m = AffineMap.transpose_map()
    assert m == AffineMap(2, 0, (AffineExpr.dimension(1), AffineExpr.dimension(0)))
    m = AffineMap.empty()
    assert m == AffineMap(0, 0, ())


def test_from_callable():
    assert AffineMap.from_callable(lambda: (1,)) == AffineMap.constant_map(1)
    assert AffineMap.from_callable(lambda: (0, 1)) == AffineMap.point_map(0, 1)
    assert AffineMap.from_callable(lambda i, j: (i, j)) == AffineMap.identity(2)
    assert AffineMap.from_callable(lambda i, j: (j, i)) == AffineMap.transpose_map()
    assert AffineMap.from_callable(lambda: ()) == AffineMap.empty()

    assert AffineMap.from_callable(
        lambda i, j, p, q: (p + i, q + j), dim_symbol_split=(2, 2)
    ) == AffineMap(
        2,
        2,
        (
            AffineBinaryOpExpr(
                AffineBinaryOpKind.Add, AffineExpr.symbol(0), AffineExpr.dimension(0)
            ),
            AffineBinaryOpExpr(
                AffineBinaryOpKind.Add, AffineExpr.symbol(1), AffineExpr.dimension(1)
            ),
        ),
    )


def test_from_callable_fail():
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Argument count mismatch in AffineMap.from_callable: 1 != 1 + 1"
        ),
    ):
        AffineMap.from_callable(lambda i: (i,), dim_symbol_split=(1, 1))


def test_inverse_permutation():
    assert AffineMap.empty().inverse_permutation() == AffineMap.empty()
    assert AffineMap.from_callable(
        lambda d0, d1, d2: (d1, d1, d0, d2, d1, d2, d1, d0)
    ).inverse_permutation() == AffineMap(
        8, 0, tuple(AffineExpr.dimension(d) for d in (2, 0, 3))
    )


@pytest.mark.parametrize(
    "input_map, expected_map",
    [
        (AffineMap.empty(), AffineMap.empty()),
        (
            AffineMap.from_callable(lambda d0, d1, d2: (d2, d1, d0)),
            AffineMap.from_callable(lambda d0, d1, d2: (d2, d1, d0)),
        ),
        (
            AffineMap.from_callable(lambda d0, d1, d2: (d1, d0)),
            AffineMap.from_callable(lambda d0, d1: (d1, d0, 0)),
        ),
        (
            AffineMap.from_callable(lambda d0, d1, d2: (d1, 0, d0)),
            AffineMap.from_callable(lambda d0, d1, d2: (d2, d0, 0)),
        ),
    ],
)
def test_inverse_and_broadcast_projected_permutation(
    input_map: AffineMap, expected_map: AffineMap
):
    inverse = input_map.inverse_and_broadcast_projected_permutation()
    assert inverse == expected_map, f"{input_map} {expected_map}"
    if inverse.is_projected_permutation():
        assert inverse.inverse_and_broadcast_projected_permutation() == input_map


t = True
f = False


@pytest.mark.parametrize(
    "input_map, drop_dims_mask, expected_map",
    [
        (
            AffineMap.from_callable(lambda d0, d1, d2: (d1, d2)),
            [t, f, f],
            AffineMap.from_callable(lambda d0, d1: (d0, d1)),
        ),
        (
            AffineMap.from_callable(lambda d0, d1, d2: (d2, d2)),
            [f, t, f],
            AffineMap.from_callable(lambda d0, d1: (d1, d1)),
        ),
    ],
)
def test_drop_dims(
    input_map: AffineMap, drop_dims_mask: Sequence[bool], expected_map: AffineMap
):
    assert input_map.drop_dims(drop_dims_mask) == expected_map


@pytest.mark.parametrize(
    "input_map, drop_results_mask, expected_map",
    [
        (
            AffineMap.from_callable(lambda d0, d1, d2: (d1, d2)),
            [t, f],
            AffineMap.from_callable(lambda d0, d1, d2: (d2,)),
        ),
        (
            AffineMap.from_callable(lambda d0, d1, d2: (d1, d2)),
            [f, t],
            AffineMap.from_callable(lambda d0, d1, d2: (d1,)),
        ),
        (
            AffineMap.from_callable(
                lambda d0, d1, s0: (d0 + s0, d1), dim_symbol_split=(2, 1)
            ),
            [t, f],
            AffineMap.from_callable(lambda d0, d1, s0: (d1,), dim_symbol_split=(2, 1)),
        ),
        (
            AffineMap.from_callable(
                lambda d0, d1, s0: (d0 + s0, d1), dim_symbol_split=(2, 1)
            ),
            [f, t],
            AffineMap.from_callable(
                lambda d0, d1, s0: (d0 + s0,), dim_symbol_split=(2, 1)
            ),
        ),
    ],
)
def test_drop_results(
    input_map: AffineMap, drop_results_mask: Sequence[bool], expected_map: AffineMap
):
    assert input_map.drop_results(drop_results_mask) == expected_map


def test_affine_expr_affine_expr_binary_simplification():
    one = AffineExpr.constant(1)
    two = AffineExpr.constant(2)
    three = AffineExpr.constant(3)
    five = AffineExpr.constant(5)
    six = AffineExpr.constant(6)

    # Should return AffineConstExpr when both lhs and rhs are AffineConstantExpr
    assert AffineExpr.binary(AffineBinaryOpKind.Add, one, one) == two
    assert AffineExpr.binary(AffineBinaryOpKind.Mul, two, three) == six
    assert AffineExpr.binary(AffineBinaryOpKind.Mod, five, two) == one
    assert AffineExpr.binary(AffineBinaryOpKind.FloorDiv, five, two) == two
    assert AffineExpr.binary(AffineBinaryOpKind.CeilDiv, five, two) == three


def test_affine_expr_used_dims():
    assert AffineExpr.dimension(1).used_dims() == {1}
    assert (AffineExpr.dimension(2) + AffineExpr.dimension(3)).used_dims() == {2, 3}
    assert AffineExpr.symbol(4).used_dims() == set()
    assert AffineExpr.constant(5).used_dims() == set()


@pytest.mark.parametrize(
    "affine_map, expected_used, expected_unused",
    [
        (
            AffineMap.from_callable(lambda i, j: (i, j)),
            {0, 1},
            set[int](),
        ),
        (
            AffineMap.from_callable(lambda i, j, _: (i + j,)),
            {0, 1},
            {2},
        ),
        (
            AffineMap.from_callable(lambda i, _, k: (i, k)),
            {0, 2},
            {1},
        ),
    ],
)
def test_affine_map_used_and_unused_dims(
    affine_map: AffineMap, expected_used: set[int], expected_unused: set[int]
):
    assert affine_map.used_dims() == expected_used
    assert affine_map.unused_dims() == expected_unused


@pytest.mark.parametrize(
    "affine_map, expected_used",
    [
        (
            AffineMap.from_callable(lambda i, j: (i, j)),
            (True, True),
        ),
        (
            AffineMap.from_callable(lambda i, j, _: (i + j,)),
            (True, True, False),
        ),
        (
            AffineMap.from_callable(lambda i, _, k: (i, k)),
            (True, False, True),
        ),
        (
            AffineMap.from_callable(lambda _, __: ()),
            (False, False),
        ),
    ],
)
def test_affine_map_used_dims_bit_vector(
    affine_map: "AffineMap", expected_used: tuple[bool, ...]
) -> None:
    expected_unused = tuple(not d for d in expected_used)
    assert affine_map.used_dims_bit_vector() == expected_used
    assert affine_map.unused_dims_bit_vector() == expected_unused


def test_minor_identity():
    assert AffineMap.empty().is_minor_identity()
    assert AffineMap.identity(3).is_minor_identity()
    assert AffineMap.minor_identity(3, 2).is_minor_identity()
    assert AffineMap.minor_identity(5, 3).is_minor_identity()

    # Test the actual structure of minor identity maps
    minor_id_3_2 = AffineMap.minor_identity(3, 2)
    assert minor_id_3_2 == AffineMap.from_callable(lambda _, d1, d2: (d1, d2))

    minor_id_5_3 = AffineMap.minor_identity(5, 3)
    assert minor_id_5_3 == AffineMap.from_callable(
        lambda _d0, _d1, d2, d3, d4: (d2, d3, d4)
    )
    # Test non-minor identity maps
    non_minor_id = AffineMap(2, 0, (AffineExpr.dimension(0), AffineExpr.dimension(0)))
    assert not non_minor_id.is_minor_identity()

    # Map with symbols is not a minor identity
    map_with_symbols = AffineMap(
        2, 1, (AffineExpr.dimension(0), AffineExpr.dimension(1))
    )
    assert not map_with_symbols.is_minor_identity()

    # Map with non-consecutive dimensions is not a minor identity
    non_consecutive = AffineMap(
        3, 0, (AffineExpr.dimension(0), AffineExpr.dimension(2))
    )
    assert not non_consecutive.is_minor_identity()

    with pytest.raises(
        ValueError,
        match="Dimension mismatch, expected dims 2 to be greater than or equal to "
        "results 3",
    ):
        AffineMap.minor_identity(2, 3)


def test_is_projected_permutation():
    assert AffineMap(0, 0, ()).is_projected_permutation()
    assert not AffineMap(0, 1, ()).is_projected_permutation()

    assert AffineMap.from_callable(lambda d0, d1: (d0, d1)).is_projected_permutation()
    assert not AffineMap.from_callable(
        lambda d0, d1: (d0, d0)
    ).is_projected_permutation()
    assert not AffineMap.from_callable(lambda d0: (d0, d0)).is_projected_permutation()

    assert AffineMap.from_callable(
        lambda d0, d1, d2: (d1, d0)
    ).is_projected_permutation()
    assert not AffineMap.from_callable(
        lambda d0, d1, d2: (d1, 0, d0)
    ).is_projected_permutation()
    assert AffineMap.from_callable(
        lambda d0, d1, d2: (d1, 0, d0)
    ).is_projected_permutation(allow_zero_in_results=True)


_T = TypeVar("_T")


@pytest.mark.parametrize(
    "permutation_map, inputs, expected",
    [
        (
            AffineMap.from_callable(lambda d0, d1, d2: (d1, d0)),
            (10, 20, 30),
            (20, 10),
        ),
        (
            AffineMap(
                8, 0, tuple(AffineExpr.dimension(d) for d in (4, 1, 6, 0, 3, 2, 5, 7))
            ),
            "policemr",
            ("c", "o", "m", "p", "i", "l", "e", "r"),
        ),
    ],
)
def test_apply_permutation_map(
    permutation_map: AffineMap, inputs: Sequence[_T], expected: tuple[_T, ...]
):
    assert permutation_map.apply_permutation(inputs) == expected


d0 = AffineExpr.dimension(0)
d1 = AffineExpr.dimension(1)
s0 = AffineExpr.symbol(0)
c2 = AffineExpr.constant(2)
c3 = AffineExpr.constant(3)


@pytest.mark.parametrize(
    "expr, expected",
    [
        # Pure affine: only addition, multiplication by constant, dimensions, symbols, constants
        (d0 + d1, True),  # d0 + d1 is pure affine
        (c2 * d0 + c3 * d1, True),  # 2 * d0 + 3 * d1 is pure affine
        (d0 * 2 + d1 * 3, True),  # d0 * 2 + d1 * 3 is pure affine
        (d0 + s0, True),  # d0 + s0 is pure affine
        (d0 + 5, True),  # d0 + 5 is pure affine
        (d0 // 2, True),  # d0 // 2 is pure affine (floordiv by constant)
        ((d0 + d1) // 3, True),  # (d0 + d1) // 3 is pure affine
        (d0 % 4, True),  # d0 % 4 is pure affine (mod by constant)
        ((d0 + d1) % 5, True),  # (d0 + d1) % 5 is pure affine
        (d0.ceil_div(7), True),  # d0.ceil_div(7) is pure affine (ceildiv by constant)
        ((d0 + d1).ceil_div(2), True),  # (d0 + d1).ceil_div(2) is pure affine
        # Not pure affine: multiplication by symbol
        (AffineBinaryOpExpr(AffineBinaryOpKind.Mul, d0, s0), False),
        # Not pure affine: multiplication by dimension
        (AffineBinaryOpExpr(AffineBinaryOpKind.Mul, d0, d1), False),
        # Not pure affine: floordiv by symbol
        (AffineBinaryOpExpr(AffineBinaryOpKind.FloorDiv, d0, s0), False),
        # Not pure affine: mod by symbol
        (AffineBinaryOpExpr(AffineBinaryOpKind.Mod, d0, s0), False),
        # Not pure affine: ceildiv by symbol
        (AffineBinaryOpExpr(AffineBinaryOpKind.CeilDiv, d0, s0), False),
    ],
)
def test_is_pure_affine(expr: AffineExpr, expected: bool):
    assert expr.is_pure_affine() is expected


def test_traversal():
    #    6
    #  4 + 5
    # 0+1 2+3
    nodes = [AffineExpr.dimension(i) for i in range(4)]
    nodes.append(nodes[0] + nodes[1])
    nodes.append(nodes[2] + nodes[3])
    nodes.append(nodes[4] + nodes[5])

    root = nodes[-1]

    dfs = tuple(root.dfs())
    assert len(dfs) == 7
    dfs_indices = [6, 4, 0, 1, 5, 2, 3]
    for i, node in zip(dfs_indices, dfs):
        assert nodes[i] is node

    post_order = tuple(root.post_order())
    post_order_indices = [0, 1, 4, 2, 3, 5, 6]
    assert len(post_order) == 7
    for i, node in zip(post_order_indices, post_order):
        assert nodes[i] is node


d0, d1, d2 = (AffineExpr.dimension(i) for i in range(3))
s0, s1, s2 = (AffineExpr.symbol(i) for i in range(3))


def test_from_flat_form():
    # Empty case
    assert AffineExpr.from_flat_form([0], 0, 0, []) == AffineExpr.constant(0)
    # One dimension
    assert AffineExpr.from_flat_form([1, 0], 1, 0, []) == d0
    # One symbol
    assert AffineExpr.from_flat_form([1, 0], 0, 1, []) == s0
    # Factors
    assert (
        AffineExpr.from_flat_form(
            [
                0,  # d0 * 0
                1,  # d1 * 1
                2,  # d2 * 2
                0,  # s0 * 0
                1,  # s1 * 1
                2,  # s2 * 2
                0,  # d0 * 0
                2,  # s0 * 2
                1,  # c1
            ],
            3,
            3,
            [d0, s0],
        )
        == d1 + d2 * 2 + s1 + s2 * 2 + s0 * 2 + 1
    )


@pytest.mark.parametrize(
    "expr,expected",
    [
        # Simple cases: constants, dimensions, symbols
        (AffineExpr.constant(0), AffineExpr.constant(0)),
        (AffineExpr.constant(42), AffineExpr.constant(42)),
        (d0, d0),
        (d1, d1),
        (s0, s0),
        (s2, s2),
        # Already-simplified sums
        (d0 + 3, d0 + 3),
        (2 * d1 + s0, 2 * d1 + s0),
        # Expressions with redundant zeros
        (d0 * 0, AffineExpr.constant(0)),
        (0 * s1, AffineExpr.constant(0)),
        (AffineExpr.constant(0) + d1, d1),
        (d0 + 0, d0),
        (0 + d2, d2),
        # Combining constants
        (
            AffineBinaryOpExpr(
                AffineBinaryOpKind.Add, AffineExpr.constant(2), AffineExpr.constant(3)
            ),
            AffineExpr.constant(5),
        ),
        (
            AffineBinaryOpExpr(
                AffineBinaryOpKind.Mul,
                AffineBinaryOpExpr(
                    AffineBinaryOpKind.Add,
                    AffineExpr.constant(1),
                    AffineExpr.constant(2),
                ),
                AffineExpr.constant(3),
            ),
            AffineExpr.constant(9),
        ),
        ((d0 + 2) + 3, d0 + 5),
        # Multiplying by 1 or 0
        (d1 * 1, d1),
        (1 * s0, s0),
        ((d1 + 2) * 1, d1 + 2),
        ((d2 + 7) * 0, AffineExpr.constant(0)),
        # Grouping like terms (if supported by simplify)
        (d0 + d0, 2 * d0),
        (2 * d0 + 3 * d0, 5 * d0),
        (d0 + s1 + d0, 2 * d0 + s1),
        # More complex: terms cancel
        (d0 - d0, AffineExpr.constant(0)),
        (2 * d1 - d1, d1),
        ((d0 + 2) - 2, d0),
        # Deeply nested expressions
        ((d0 + (AffineExpr.constant(1) + 2)), d0 + 3),
        (((d0 + 2) + (d0 + 3)), 2 * d0 + 5),
        # To be added in upcoming PRs adding div and mod support
        # # Exact multiples of floordiv are cancelled out
        # ((2 * d0 + 4 * s1) // 2, d0 + 2 * s1),
        # # Exact multiples of ceildiv are cancelled out
        # ((2 * d0 + 4 * s1).ceil_div(2), d0 + 2 * s1),
        # # Greatest common denominators of floordiv are cancelled out
        # ((2 * d0 + 4 * s1) // 6, (d0 + 2 * s1) // 3),
        # # Greatest common denominators of ceildiv are cancelled out
        # ((2 * d0 + 4 * s1).ceil_div(6), (d0 + 2 * s1).ceil_div(3)),
        # # Division is aggregated
        # (d0 // 2 + d0 // 2, d0 // 2 * 2),
        # # Factors and modulo cancel
        # ((2 * d0 + 4 * s1) % 2, AffineExpr.constant(0)),
        # # Modulo is converted into the manual computation (x % k == x - (x // k * k))
        # ((2 * d0 + 4 * s1) % 3, (2 * d0 + 4 * s1) + ((2 * d0 + 4 * s1) // 3 * -3)),
        # # This sometimes cancels out
        # (d0 % 3 + d0 // 3 * 3, d0),
        # # GCD cancel in the fraction
        # ((2 * d0 + 4 * s1) % 6, (2 * d0 + 4 * s1) + ((d0 + 2 * s1) // 3 * -6)),
        # # Modulos with same expressions are aggregated
        # (d0 % 2 - d0 % 2, AffineExpr.constant(0)),
    ],
)
def test_affine_expr_simplify(expr: AffineExpr, expected: AffineExpr):
    assert str(expr.simplify(3, 3)) == str(expected)
    assert expr.simplify(3, 3) == expected
    # Eval on some numbers to catch errors in simplification test
    assert expr.eval((1, 2, 3), (5, 7, 11)) == expected.eval((1, 2, 3), (5, 7, 11))
    assert expr.eval((11, 7, 5), (3, 2, 1)) == expected.eval((11, 7, 5), (3, 2, 1))


@pytest.mark.parametrize(
    "expr",
    [
        (lambda: AffineBinaryOpExpr(AffineBinaryOpKind.Mul, d0, s0)),
        # To be added in upcoming PRs adding div and mod support
        # (lambda: AffineBinaryOpExpr(AffineBinaryOpKind.FloorDiv, d0, s0)),
        # (lambda: AffineBinaryOpExpr(AffineBinaryOpKind.Mod, d0, s0)),
    ],
)
def test_affine_expr_simplify_semi_affine_raises(expr: Callable[[], AffineExpr]):
    with pytest.raises(
        NotImplementedError,
        match="Simplification of semi-affine expressions is not implemented yet.",
    ):
        expr().simplify(num_dims=3, num_symbols=3)
    with pytest.raises(
        NotImplementedError,
        match="Semi-affine map flattening not implemented",
    ):
        SimpleAffineExprFlattener(3, 3).simplify(expr())



import pytest

from xdsl.dialects.builtin import ModuleOp, StringAttr
from xdsl.dialects.test import TestOp
from xdsl.ir import Block, Region
from xdsl.utils.symbol_table import (
    SymbolTable,
    SymbolTableCollection,
    Visibility,
    walk_symbol_table,
)


def test_walk_symbol_table():
    a = TestOp(
        regions=[
            Region(
                Block(
                    [
                        TestOp(properties={"sym_name": StringAttr("aa")}),
                        ModuleOp(
                            [TestOp(properties={"sym_name": StringAttr("aba")})],
                            sym_name=StringAttr("ab"),
                        ),
                        TestOp(properties={"sym_name": StringAttr("ac")}),
                    ]
                )
            )
        ],
        properties={"sym_name": StringAttr("a")},
    )

    sym_names = [
        getattr(op.properties["sym_name"], "data") for op in walk_symbol_table(a)
    ]

    unique_sym_names = set(sym_names)

    assert unique_sym_names == {"a", "aa", "ab", "ac"}
    assert len(unique_sym_names) == len(sym_names)


# SymbolTable instance method tests
def test_symbol_table_lookup():
    """Test SymbolTable.lookup method."""
    module = ModuleOp([], sym_name=StringAttr("test_module"))
    symbol_table = SymbolTable(module)

    # This will raise NotImplementedError until implemented
    with pytest.raises(NotImplementedError):
        symbol_table.lookup("test_symbol")


def test_symbol_table_remove():
    """Test SymbolTable.remove method."""
    module = ModuleOp([], sym_name=StringAttr("test_module"))
    symbol_table = SymbolTable(module)
    test_op = TestOp()

    # This will raise NotImplementedError until implemented
    with pytest.raises(NotImplementedError):
        symbol_table.remove(test_op)


def test_symbol_table_erase():
    """Test SymbolTable.erase method."""
    module = ModuleOp([], sym_name=StringAttr("test_module"))
    symbol_table = SymbolTable(module)
    test_op = TestOp()

    # This will raise NotImplementedError until implemented
    with pytest.raises(NotImplementedError):
        symbol_table.erase(test_op)


def test_symbol_table_insert():
    """Test SymbolTable.insert method."""
    from xdsl.builder import InsertPoint

    module = ModuleOp([], sym_name=StringAttr("test_module"))
    symbol_table = SymbolTable(module)
    test_op = TestOp()
    insertion_point = InsertPoint(module.regions[0].blocks[0])

    # This will raise NotImplementedError until implemented
    with pytest.raises(NotImplementedError):
        symbol_table.insert(test_op, insertion_point)


def test_symbol_table_rename():
    """Test SymbolTable.rename method."""
    module = ModuleOp([], sym_name=StringAttr("test_module"))
    symbol_table = SymbolTable(module)
    test_op = TestOp()

    # This will raise NotImplementedError until implemented
    with pytest.raises(NotImplementedError):
        symbol_table.rename(test_op, "new_name")


def test_symbol_table_rename_to_unique():
    """Test SymbolTable.rename_to_unique method."""
    module = ModuleOp([], sym_name=StringAttr("test_module"))
    symbol_table = SymbolTable(module)
    test_op = TestOp()
    other_tables: list[SymbolTable] = []

    # This will raise NotImplementedError until implemented
    with pytest.raises(NotImplementedError):
        symbol_table.rename_to_unique(test_op, other_tables)


# SymbolTable static method tests
def test_symbol_table_get_symbol_name():
    """Test SymbolTable.get_symbol_name static method."""
    test_op = TestOp()

    # This will raise NotImplementedError until implemented
    with pytest.raises(NotImplementedError):
        SymbolTable.get_symbol_name(test_op)


def test_symbol_table_set_symbol_name():
    """Test SymbolTable.set_symbol_name static method."""
    test_op = TestOp()
    name = StringAttr("test_name")

    # This will raise NotImplementedError until implemented
    with pytest.raises(NotImplementedError):
        SymbolTable.set_symbol_name(test_op, name)


def test_symbol_table_get_symbol_visibility():
    """Test SymbolTable.get_symbol_visibility static method."""
    test_op = TestOp()

    # This will raise NotImplementedError until implemented
    with pytest.raises(NotImplementedError):
        SymbolTable.get_symbol_visibility(test_op)


def test_symbol_table_set_symbol_visibility():
    """Test SymbolTable.set_symbol_visibility static method."""
    test_op = TestOp()
    visibility = Visibility.PUBLIC

    # This will raise NotImplementedError until implemented
    with pytest.raises(NotImplementedError):
        SymbolTable.set_symbol_visibility(test_op, visibility)


def test_symbol_table_get_nearest_symbol_table():
    """Test SymbolTable.get_nearest_symbol_table static method."""
    test_op = TestOp()

    # This will raise NotImplementedError until implemented
    with pytest.raises(NotImplementedError):
        SymbolTable.get_nearest_symbol_table(test_op)


def test_symbol_table_walk_symbol_tables():
    """Test SymbolTable.walk_symbol_tables static method."""
    test_op = TestOp()

    # This will raise NotImplementedError until implemented
    with pytest.raises(NotImplementedError):
        list(SymbolTable.walk_symbol_tables(test_op, True))


def test_symbol_table_lookup_symbol_in():
    """Test SymbolTable.lookup_symbol_in static method."""
    test_op = TestOp()
    symbol = StringAttr("test_symbol")

    # This will raise NotImplementedError until implemented
    with pytest.raises(NotImplementedError):
        SymbolTable.lookup_symbol_in(test_op, symbol, all_symbols=False)


def test_symbol_table_lookup_nearest_symbol_from():
    """Test SymbolTable.lookup_nearest_symbol_from static method."""
    test_op = TestOp()
    symbol = StringAttr("test_symbol")

    # This will raise NotImplementedError until implemented
    with pytest.raises(NotImplementedError):
        SymbolTable.lookup_nearest_symbol_from(test_op, symbol)


def test_symbol_table_get_symbol_uses():
    """Test SymbolTable.get_symbol_uses static method."""
    test_op = TestOp()

    # This will raise NotImplementedError until implemented
    with pytest.raises(NotImplementedError):
        SymbolTable.get_symbol_uses(test_op)


def test_symbol_table_symbol_known_use_empty():
    """Test SymbolTable.symbol_known_use_empty static method."""
    test_op = TestOp()
    symbol = StringAttr("test_symbol")

    # This will raise NotImplementedError until implemented
    with pytest.raises(NotImplementedError):
        SymbolTable.symbol_known_use_empty(symbol, test_op)


def test_symbol_table_replace_all_symbol_uses():
    """Test SymbolTable.replace_all_symbol_uses static method."""
    test_op = TestOp()
    old_symbol = StringAttr("old_symbol")
    new_symbol = StringAttr("new_symbol")

    # This will raise NotImplementedError until implemented
    with pytest.raises(NotImplementedError):
        SymbolTable.replace_all_symbol_uses(old_symbol, new_symbol, test_op)


# SymbolTableCollection tests
def test_symbol_table_collection_init():
    """Test SymbolTableCollection initialization."""
    collection = SymbolTableCollection()
    assert collection.symbol_tables == {}


def test_symbol_table_collection_lookup_symbol_in():
    """Test SymbolTableCollection.lookup_symbol_in static method."""
    test_op = TestOp()
    symbol = StringAttr("test_symbol")

    # This will raise NotImplementedError until implemented
    with pytest.raises(NotImplementedError):
        SymbolTableCollection.lookup_symbol_in(test_op, symbol, all_symbols=False)


def test_symbol_table_collection_lookup_nearest_symbol_from():
    """Test SymbolTableCollection.lookup_nearest_symbol_from static method."""
    test_op = TestOp()
    symbol = StringAttr("test_symbol")

    # This will raise NotImplementedError until implemented
    with pytest.raises(NotImplementedError):
        SymbolTableCollection.lookup_nearest_symbol_from(test_op, symbol)


def test_symbol_table_collection_get_symbol_table():
    """Test SymbolTableCollection.get_symbol_table method."""
    collection = SymbolTableCollection()
    test_op = TestOp()

    # This will raise NotImplementedError until implemented
    with pytest.raises(NotImplementedError):
        collection.get_symbol_table(test_op)



# RUN: python %s | filecheck %s

from collections.abc import Callable
from ctypes import c_size_t

from xdsl.dialects import builtin
from xdsl.frontend.pyast.context import PyASTContext
from xdsl.frontend.pyast.utils.exceptions import CodeGenerationException

ctx = PyASTContext(post_transforms=[])
ctx.register_type(c_size_t, builtin.IndexType())


# CHECK: For loops are currently not supported!
@ctx.parse_program
def test_for_unsupported(end: c_size_t):
    for _ in range(
        end  # pyright: ignore[reportArgumentType]
    ):
        pass
    return


try:
    test_for_unsupported.module
except NotImplementedError as e:
    print(e)


# CHECK: Unsupported function argument type: 'Callable[..., None]'
@ctx.parse_program
def test_complex_arg_annotation(x: Callable[..., None]) -> None:
    return


try:
    test_complex_arg_annotation.module
except CodeGenerationException as e:
    print(e.msg)


# CHECK: Unsupported function return type: 'int | None'
@ctx.parse_program
def test_complex_return_annotation() -> int | None:
    return


try:
    test_complex_return_annotation.module
except CodeGenerationException as e:
    print(e.msg)



"""
xDSL leverages Python [plugins](https://packaging.python.org/en/latest/guides/creating-and-discovering-plugins/) for user-provided compiler infrastructure.

Python packages that want to extend the passes and dialects available in `xdsl-opt` can
implement their own `Universe`, which will be discovered by `Multiverse` at runtime.

To opt into this behavior first add a Universe instance somewhere in your project,
for example in `your_project/universe.py`:

```py
def get_dialect():
    ...

def get_pass():
    ...

YOUR_UNIVERSE = Universe(
    all_dialects={"your_dialect": get_dialect},
    all_passes={"your-pass": get_pass},
)
```

Then, add the following entry in your `pyproject.toml` file:

```toml
[project.entry-points.'xdsl.universe']
your_project = 'your_project.universe:YOUR_UNIVERSE'
```

Note that any clashes in names of dialects or passes will result in an error.
"""  # noqa: E501

from __future__ import annotations

from collections.abc import Callable

from xdsl.dialects import get_all_dialects
from xdsl.ir import Dialect
from xdsl.passes import ModulePass
from xdsl.transforms import get_all_passes


class Universe:
    """
    Contains all the dialects and passes from modules in the environment.
    """

    all_dialects: dict[str, Callable[[], Dialect]]
    """Dialects from all modules in the current environment."""
    all_passes: dict[str, Callable[[], type[ModulePass]]]
    """Passes from all modules in the current environment."""

    def __init__(
        self,
        *,
        all_dialects: dict[str, Callable[[], Dialect]] | None = None,
        all_passes: dict[str, Callable[[], type[ModulePass]]] | None = None,
    ) -> None:
        self.all_dialects = {} if all_dialects is None else all_dialects
        self.all_passes = {} if all_passes is None else all_passes

    @staticmethod
    def get_multiverse() -> Universe:
        """
        Traverses the current environment looking for entry points in the
        `xdsl.universe` group.
        """
        from importlib.metadata import entry_points

        discovered_plugins = entry_points(group="xdsl.universe")
        all_universes = {plugin.name: plugin.load() for plugin in discovered_plugins}
        sorted_universes = sorted(all_universes.items(), key=lambda t: t[0])

        all_dialects: dict[str, Callable[[], Dialect]] = {}
        all_passes: dict[str, Callable[[], type[ModulePass]]] = {}

        for universe_name, universe in sorted_universes:
            if not isinstance(universe, Universe):
                raise ValueError(
                    f"Entry point {universe} for plugin {universe_name} is not an "
                    "instance of Universe."
                )
            for dialect_name, dialect in universe.all_dialects.items():
                if dialect_name in all_dialects:
                    universes = [
                        n for n, u in sorted_universes if dialect_name in u.all_dialects
                    ]
                    raise ValueError(
                        f"Duplicate definition of {dialect_name} in {universes}."
                    )
                all_dialects[dialect_name] = dialect

            for pass_name, _pass in universe.all_passes.items():
                if pass_name in all_passes:
                    universes = [
                        n for n, u in sorted_universes if pass_name in u.all_passes
                    ]
                    raise ValueError(
                        f"Duplicate definition of {pass_name} in {universes}."
                    )
                all_passes[pass_name] = _pass

        return Universe(all_dialects=all_dialects, all_passes=all_passes)


XDSL_UNIVERSE = Universe(all_dialects=get_all_dialects(), all_passes=get_all_passes())



import pytest

from xdsl.dialects import x86
from xdsl.dialects.builtin import IntegerAttr, i32
from xdsl.dialects.x86.ops import (
    DM_LeaOp,
    DM_MovOp,
    DM_VbroadcastsdOp,
    DM_VbroadcastssOp,
    DM_VmovupdOp,
    DM_VmovupsOp,
    DMI_ImulOp,
    M_DecOp,
    M_IDivOp,
    M_ImulOp,
    M_IncOp,
    M_NegOp,
    M_NotOp,
    MI_AddOp,
    MI_AndOp,
    MI_CmpOp,
    MI_MovOp,
    MI_OrOp,
    MI_SubOp,
    MI_XorOp,
    MS_AddOp,
    MS_AndOp,
    MS_CmpOp,
    MS_MovOp,
    MS_OrOp,
    MS_SubOp,
    MS_XorOp,
    RM_AddOp,
    RM_AndOp,
    RM_ImulOp,
    RM_OrOp,
    RM_SubOp,
    RM_XorOp,
    SM_CmpOp,
)
from xdsl.ir import Block, Operation
from xdsl.traits import MemoryReadEffect
from xdsl.transforms.canonicalization_patterns.x86 import get_constant_value
from xdsl.utils.test_value import create_ssa_value


def test_unallocated_register():
    unallocated = x86.registers.GeneralRegisterType.from_name("")
    assert not unallocated.is_allocated
    assert unallocated == x86.registers.UNALLOCATED_GENERAL

    unallocated = x86.registers.RFLAGSRegisterType.from_name("")
    assert not unallocated.is_allocated
    assert unallocated == x86.registers.UNALLOCATED_RFLAGS

    unallocated = x86.registers.AVX2RegisterType.from_name("")
    assert not unallocated.is_allocated
    assert unallocated == x86.registers.UNALLOCATED_AVX2

    unallocated = x86.registers.AVX512RegisterType.from_name("")
    assert not unallocated.is_allocated
    assert unallocated == x86.registers.UNALLOCATED_AVX512

    unallocated = x86.registers.SSERegisterType.from_name("")
    assert not unallocated.is_allocated
    assert unallocated == x86.registers.UNALLOCATED_SSE


@pytest.mark.parametrize(
    "register, name",
    [
        (x86.registers.RAX, "rax"),
        (x86.registers.RCX, "rcx"),
        (x86.registers.RDX, "rdx"),
        (x86.registers.RBX, "rbx"),
        (x86.registers.RSP, "rsp"),
        (x86.registers.RBP, "rbp"),
        (x86.registers.RSI, "rsi"),
        (x86.registers.RDI, "rdi"),
        # Currently don't support 32-bit registers
        # https://github.com/xdslproject/xdsl/issues/4737
        # (x86.register.EAX, "eax"),
        # (x86.register.ECX, "ecx"),
        # (x86.register.EDX, "edx"),
        # (x86.register.EBX, "ebx"),
        # (x86.register.ESP, "esp"),
        # (x86.register.EBP, "ebp"),
        # (x86.register.ESI, "esi"),
        # (x86.register.EDI, "edi"),
        (x86.registers.R8, "r8"),
        (x86.registers.R9, "r9"),
        (x86.registers.R10, "r10"),
        (x86.registers.R11, "r11"),
        (x86.registers.R12, "r12"),
        (x86.registers.R13, "r13"),
        (x86.registers.R14, "r14"),
        (x86.registers.R15, "r15"),
    ],
)
def test_register(register: x86.registers.GeneralRegisterType, name: str):
    assert register.is_allocated
    assert register.register_name.data == name


def test_rflags_register():
    rflags = x86.registers.RFLAGS
    assert rflags.is_allocated
    assert rflags.register_name.data == "rflags"


@pytest.mark.parametrize(
    "register, name",
    [
        (x86.registers.ZMM0, "zmm0"),
        (x86.registers.ZMM1, "zmm1"),
        (x86.registers.ZMM2, "zmm2"),
        (x86.registers.ZMM3, "zmm3"),
        (x86.registers.ZMM4, "zmm4"),
        (x86.registers.ZMM5, "zmm5"),
        (x86.registers.ZMM6, "zmm6"),
        (x86.registers.ZMM7, "zmm7"),
        (x86.registers.ZMM8, "zmm8"),
        (x86.registers.ZMM9, "zmm9"),
        (x86.registers.ZMM10, "zmm10"),
        (x86.registers.ZMM11, "zmm11"),
        (x86.registers.ZMM12, "zmm12"),
        (x86.registers.ZMM13, "zmm13"),
        (x86.registers.ZMM14, "zmm14"),
        (x86.registers.ZMM15, "zmm15"),
        (x86.registers.ZMM16, "zmm16"),
        (x86.registers.ZMM17, "zmm17"),
        (x86.registers.ZMM18, "zmm18"),
        (x86.registers.ZMM19, "zmm19"),
        (x86.registers.ZMM20, "zmm20"),
        (x86.registers.ZMM21, "zmm21"),
        (x86.registers.ZMM22, "zmm22"),
        (x86.registers.ZMM23, "zmm23"),
        (x86.registers.ZMM24, "zmm24"),
        (x86.registers.ZMM25, "zmm25"),
        (x86.registers.ZMM26, "zmm26"),
        (x86.registers.ZMM27, "zmm27"),
        (x86.registers.ZMM28, "zmm28"),
        (x86.registers.ZMM29, "zmm29"),
        (x86.registers.ZMM30, "zmm30"),
        (x86.registers.ZMM31, "zmm31"),
    ],
)
def test_avx512_register(register: x86.registers.AVX512RegisterType, name: str):
    assert register.is_allocated
    assert register.register_name.data == name


@pytest.mark.parametrize(
    "register, name",
    [
        (x86.registers.YMM0, "ymm0"),
        (x86.registers.YMM1, "ymm1"),
        (x86.registers.YMM2, "ymm2"),
        (x86.registers.YMM3, "ymm3"),
        (x86.registers.YMM4, "ymm4"),
        (x86.registers.YMM5, "ymm5"),
        (x86.registers.YMM6, "ymm6"),
        (x86.registers.YMM7, "ymm7"),
        (x86.registers.YMM8, "ymm8"),
        (x86.registers.YMM9, "ymm9"),
        (x86.registers.YMM10, "ymm10"),
        (x86.registers.YMM11, "ymm11"),
        (x86.registers.YMM12, "ymm12"),
        (x86.registers.YMM13, "ymm13"),
        (x86.registers.YMM14, "ymm14"),
        (x86.registers.YMM15, "ymm15"),
    ],
)
def test_avx2_register(register: x86.registers.AVX2RegisterType, name: str):
    assert register.is_allocated
    assert register.register_name.data == name


@pytest.mark.parametrize(
    "register, name",
    [
        (x86.registers.XMM0, "xmm0"),
        (x86.registers.XMM1, "xmm1"),
        (x86.registers.XMM2, "xmm2"),
        (x86.registers.XMM3, "xmm3"),
        (x86.registers.XMM4, "xmm4"),
        (x86.registers.XMM5, "xmm5"),
        (x86.registers.XMM6, "xmm6"),
        (x86.registers.XMM7, "xmm7"),
        (x86.registers.XMM8, "xmm8"),
        (x86.registers.XMM9, "xmm9"),
        (x86.registers.XMM10, "xmm10"),
        (x86.registers.XMM11, "xmm11"),
        (x86.registers.XMM12, "xmm12"),
        (x86.registers.XMM13, "xmm13"),
        (x86.registers.XMM14, "xmm14"),
        (x86.registers.XMM15, "xmm15"),
    ],
)
def test_sse_register(register: x86.registers.SSERegisterType, name: str):
    assert register.is_allocated
    assert register.register_name.data == name


@pytest.mark.parametrize(
    "OpClass, dest, operand1, operand2",
    [
        (
            x86.ops.RSS_Vfmadd231pdOp,
            x86.registers.YMM0,
            x86.registers.YMM1,
            x86.registers.YMM2,
        ),
        (
            x86.ops.RSS_Vfmadd231psOp,
            x86.registers.YMM0,
            x86.registers.YMM1,
            x86.registers.YMM2,
        ),
    ],
)
def test_rrr_vops(
    OpClass: type[
        x86.ops.RSS_Operation[
            x86.registers.X86VectorRegisterType,
            x86.registers.X86VectorRegisterType,
            x86.registers.X86VectorRegisterType,
        ]
    ],
    dest: x86.registers.X86VectorRegisterType,
    operand1: x86.registers.X86VectorRegisterType,
    operand2: x86.registers.X86VectorRegisterType,
):
    output = create_ssa_value(dest)
    param1 = create_ssa_value(operand1)
    param2 = create_ssa_value(operand2)
    op = OpClass(
        source2=output,
        register_in=param1,
        source1=param2,
        register_out=dest,
    )
    assert op.register_in.type == operand1
    assert op.source1.type == operand2
    assert op.source2.type == dest


@pytest.mark.parametrize(
    "OpClass, dest, src",
    [
        (
            x86.ops.MS_VmovupsOp,
            x86.registers.RCX,
            x86.registers.YMM0,
        ),
        (
            x86.ops.MS_VmovapdOp,
            x86.registers.RCX,
            x86.registers.YMM0,
        ),
    ],
)
def test_mr_vops(
    OpClass: type[
        x86.ops.MS_Operation[
            x86.registers.GeneralRegisterType, x86.registers.X86VectorRegisterType
        ]
    ],
    dest: x86.registers.GeneralRegisterType,
    src: x86.registers.X86VectorRegisterType,
):
    output = x86.ops.GetRegisterOp(dest)
    input = x86.ops.GetAVXRegisterOp(src)
    op = OpClass(memory=output, source=input, memory_offset=IntegerAttr(0, 64))
    assert op.memory.type == dest
    assert op.source.type == src


@pytest.mark.parametrize(
    "OpClass, dest, src",
    [
        (
            x86.ops.DM_VmovupsOp,
            x86.registers.YMM0,
            x86.registers.RCX,
        ),
        (
            x86.ops.DM_VbroadcastsdOp,
            x86.registers.YMM0,
            x86.registers.RCX,
        ),
        (
            x86.ops.DM_VbroadcastssOp,
            x86.registers.YMM0,
            x86.registers.RCX,
        ),
    ],
)
def test_rm_vops(
    OpClass: type[
        x86.ops.DM_Operation[
            x86.registers.X86VectorRegisterType, x86.registers.GeneralRegisterType
        ]
    ],
    dest: x86.registers.X86VectorRegisterType,
    src: x86.registers.GeneralRegisterType,
):
    input = x86.ops.GetRegisterOp(src)
    op = OpClass(memory=input, destination=dest, memory_offset=IntegerAttr(0, 64))
    assert op.memory.type == src
    assert op.destination.type == dest


@pytest.mark.parametrize(
    "OpClass, dest, operand1, operand2",
    [
        (
            x86.ops.DSS_AddpdOp,
            x86.registers.YMM0,
            x86.registers.YMM1,
            x86.registers.YMM2,
        ),
        (
            x86.ops.DSS_AddpsOp,
            x86.registers.YMM0,
            x86.registers.YMM1,
            x86.registers.YMM2,
        ),
    ],
)
def test_dss_vops(
    OpClass: type[
        x86.ops.DSS_Operation[
            x86.registers.X86VectorRegisterType,
            x86.registers.X86VectorRegisterType,
            x86.registers.X86VectorRegisterType,
        ]
    ],
    dest: x86.registers.X86VectorRegisterType,
    operand1: x86.registers.X86VectorRegisterType,
    operand2: x86.registers.X86VectorRegisterType,
):
    param1 = create_ssa_value(operand1)
    param2 = create_ssa_value(operand2)
    op = OpClass(param1, param2, destination=dest)
    assert op.destination.type == dest
    assert op.source1.type == operand1
    assert op.source2.type == operand2


def test_get_constant_value():
    U = x86.registers.UNALLOCATED_GENERAL
    unknown_value = create_ssa_value(U)
    assert get_constant_value(unknown_value) is None
    known_value = x86.DI_MovOp(42, destination=U).destination
    assert get_constant_value(known_value) == IntegerAttr(42, i32)
    moved_once = x86.DS_MovOp(known_value, destination=U).destination
    assert get_constant_value(moved_once) == IntegerAttr(42, i32)
    moved_twice = x86.DS_MovOp(known_value, destination=U).destination
    assert get_constant_value(moved_twice) == IntegerAttr(42, i32)

    block = Block(arg_types=(U,))
    assert get_constant_value(block.args[0]) is None


@pytest.mark.parametrize(
    "op",
    [
        DM_MovOp,
        DM_LeaOp,
        DM_VmovupsOp,
        DM_VmovupdOp,
        DM_VbroadcastsdOp,
        DM_VbroadcastssOp,
        RM_AddOp,
        RM_SubOp,
        RM_ImulOp,
        RM_AndOp,
        RM_OrOp,
        RM_XorOp,
        MS_AddOp,
        MS_SubOp,
        MS_AndOp,
        MS_OrOp,
        MS_XorOp,
        MS_MovOp,
        MI_AddOp,
        MI_SubOp,
        MI_AndOp,
        MI_OrOp,
        MI_XorOp,
        MI_MovOp,
        DMI_ImulOp,
        M_NegOp,
        M_NotOp,
        M_IncOp,
        M_DecOp,
        M_IDivOp,
        M_ImulOp,
        SM_CmpOp,
        MS_CmpOp,
        MI_CmpOp,
    ],
)
def test_read_effects(op: type[Operation]):
    assert MemoryReadEffect() in op.traits.traits


def test_jmp_numeric_label_not_implemented():
    label_op = x86.ops.LabelOp("123")
    op = x86.ops.C_JmpOp(block_values=[], successor=Block([label_op]))
    with pytest.raises(
        NotImplementedError,
        match="Assembly printing for jumps to numeric labels not implemented",
    ):
        op.assembly_line_args()
    label_op.label = x86.attributes.LabelAttr("hello")
    assert op.assembly_line_args() == ("hello",)


def test_conditional_jump_numeric_label_not_implemented():
    label_op = x86.ops.LabelOp("123")
    rflags = create_ssa_value(x86.registers.RFLAGS)
    op = x86.ops.C_JeOp(
        rflags=rflags,
        then_values=[],
        else_values=[],
        then_block=Block([label_op]),
        else_block=Block([x86.ops.LabelOp("loop_start")]),
    )
    with pytest.raises(
        NotImplementedError,
        match="Assembly printing for jumps to numeric labels not implemented",
    ):
        op.assembly_line_args()
    label_op.label = x86.attributes.LabelAttr("hello")
    assert op.assembly_line_args() == ("hello",)



import re
from dataclasses import dataclass
from typing import Generic

from typing_extensions import TypeVar

COMMENTS = r"(?:\/\/[^\n\r]+?(?:\*\)|[\n\r]))"
WHITESPACES = re.compile(r"(?:\s|" + COMMENTS + r")*")


@dataclass
class Location:
    pos: int


_LocatedCovT = TypeVar("_LocatedCovT", covariant=True)


@dataclass(frozen=True)
class Located(Generic[_LocatedCovT]):
    loc: Location
    value: _LocatedCovT

    def __bool__(self) -> bool:
        return bool(self.value)


class CodeCursor:
    code: str
    pos: int

    def __init__(self, code: str):
        self.code = code
        self.pos = 0

    def _whitespace_end(self) -> int:
        match = WHITESPACES.match(self.code, self.pos)
        assert match is not None
        return match.end()

    def skip_whitespaces(self):
        self.pos = self._whitespace_end()

    def next_regex(self, regex: re.Pattern[str]) -> Located[re.Match[str] | None]:
        match = self.peek_regex(regex)
        if match.value is not None:
            self.pos = match.value.end()
        return match

    def peek_regex(self, regex: re.Pattern[str]) -> Located[re.Match[str] | None]:
        pos = self._whitespace_end()
        return Located(Location(pos), regex.match(self.code, pos))


@dataclass
class ParseError(Exception):
    position: int
    msg: str

    @staticmethod
    def from_loc(loc: Location, msg: str) -> "ParseError":
        return ParseError(loc.pos, msg)

    def line_column(self, code: str) -> tuple[int, int]:
        if self.position < 0 or self.position > len(code):
            raise ValueError("error position is out of code range")

        line = code.count("\n", 0, self.position) + 1

        last_newline = code.rfind("\n", 0, self.position)
        column = self.position - last_newline
        return line, column



from dataclasses import dataclass, field

from xdsl.context import Context
from xdsl.dialects.builtin import ModuleOp
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import PatternRewriter
from xdsl.traits import HasCanonicalizationPatternsTrait
from xdsl.utils.op_selector import OpSelector


@dataclass(frozen=True)
class ApplyIndividualRewritePass(ModulePass):
    """
    Module pass representing the application of an individual rewrite pattern to a module.

    Matches the operation at the provided index within the module and applies the rewrite
    pattern specified by the operation and pattern names.
    """

    name = "apply-individual-rewrite"

    matched_operation_index: int = field()
    operation_name: str = field()
    pattern_name: str = field()

    def apply(self, ctx: Context, op: ModuleOp) -> None:
        matched_operation = OpSelector(
            self.matched_operation_index, self.operation_name
        ).get_op(op)
        rewriter = PatternRewriter(matched_operation)

        for trait in matched_operation.get_traits_of_type(
            HasCanonicalizationPatternsTrait
        ):
            for pattern in trait.get_canonicalization_patterns():
                if type(pattern).__name__ == self.pattern_name:
                    pattern.match_and_rewrite(matched_operation, rewriter)
                    if not rewriter.has_done_action:
                        raise ValueError(
                            f"Invalid rewrite ({self.pattern_name}) for operation "
                            f"({matched_operation}) at location "
                            f"{self.matched_operation_index}."
                        )
                    return

        raise ValueError(
            f"Pattern name {self.pattern_name} not found for the provided operation name."
        )

    @classmethod
    def schedule_space(cls, ctx: Context, module_op: ModuleOp):
        res: list[ApplyIndividualRewritePass] = []

        for op_idx, matched_op in enumerate(module_op.walk()):
            if (
                trait := matched_op.get_trait(HasCanonicalizationPatternsTrait)
            ) is None:
                continue

            pattern_by_name = {
                type(pattern).__name__: pattern
                for pattern in trait.get_canonicalization_patterns()
            }
            selector = OpSelector(op_idx, matched_op.name)

            for pattern_name, pattern in pattern_by_name.items():
                cloned_op = selector.get_op(module_op.clone())
                rewriter = PatternRewriter(cloned_op)
                pattern.match_and_rewrite(cloned_op, rewriter)
                if rewriter.has_done_action:
                    res.append(
                        ApplyIndividualRewritePass(
                            op_idx, cloned_op.name, pattern_name
                        ),
                    )

        return tuple(res)



import xdsl.frontend.listlang.list_dialect as list_dialect
from xdsl.context import Context
from xdsl.dialects import arith, builtin, func, printf, scf, tensor
from xdsl.frontend.listlang.lang_types import ListLangType
from xdsl.ir import Attribute, Block, Region
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.rewriter import InsertPoint
from xdsl.utils.hints import isa


def _name_hint_ext(name_hint: str | None, extension: str) -> str | None:
    if name_hint is None:
        return None
    if name_hint[0] != "_":
        name_hint = "_" + name_hint
    return f"{name_hint}_{extension}"


def _list_type_to_tensor(li: Attribute) -> builtin.TensorType:
    if isa(li, list_dialect.ListType):
        return builtin.TensorType(li.elem_type, (builtin.DYNAMIC_INDEX,))
    if isa(li, builtin.TensorType):
        return li
    raise ValueError("unexpected type")


class LowerLengthOp(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: list_dialect.LengthOp, rewriter: PatternRewriter):
        zero_index = rewriter.insert_op(
            arith.ConstantOp(builtin.IntegerAttr(0, builtin.IndexType()))
        )
        dim = rewriter.insert_op(tensor.DimOp(op.li, zero_index))
        cast = arith.IndexCastOp(dim, builtin.i32)
        cast.result.name_hint = op.result.name_hint
        rewriter.replace_matched_op(cast)


class LowerMapOp(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: list_dialect.MapOp, rewriter: PatternRewriter):
        zero = rewriter.insert_op(
            arith.ConstantOp(builtin.IntegerAttr(0, builtin.IndexType()))
        )
        one = rewriter.insert_op(
            arith.ConstantOp(builtin.IntegerAttr(1, builtin.IndexType()))
        )
        list_len = rewriter.insert_op(tensor.DimOp(op.li, zero))

        tensor_type = _list_type_to_tensor(op.li.type)
        result_tensor_type = _list_type_to_tensor(op.result.type)

        result_uninit = rewriter.insert_op(
            tensor.EmptyOp([list_len.result], result_tensor_type)
        )
        result_uninit.results[0].name_hint = _name_hint_ext(
            op.result.name_hint, "uninit"
        )

        rewriter_ip = rewriter.insertion_point

        for_body = Block(
            [],
            arg_types=(builtin.IndexType(), result_tensor_type),
        )
        ind_var = for_body.args[0]
        ind_var.name_hint = "_i"
        tensor_arg = for_body.args[1]

        rewriter.insertion_point = InsertPoint.at_start(for_body)

        x = rewriter.insert_op(
            tensor.ExtractOp(op.li, [ind_var], tensor_type.element_type)
        )
        x.result.name_hint = op.body.block.args[0].name_hint

        rewriter.inline_block(op.body.block, InsertPoint.at_end(for_body), (x.result,))

        closure_yield = for_body.last_op
        assert isa(closure_yield, list_dialect.YieldOp)
        result_scalar = closure_yield.yielded
        rewriter.erase_op(closure_yield)

        result = rewriter.insert_op(
            tensor.InsertOp(result_scalar, tensor_arg, [ind_var])
        )

        rewriter.insert_op(scf.YieldOp(result.result))

        rewriter.insertion_point = rewriter_ip

        for_op = scf.ForOp(
            zero,
            list_len,
            one,
            result_uninit.results,
            for_body,
        )
        for_op.results[0].name_hint = op.result.name_hint
        rewriter.replace_matched_op(for_op)


class LowerPrintOp(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: list_dialect.PrintOp, rewriter: PatternRewriter):
        zero = rewriter.insert_op(
            arith.ConstantOp(builtin.IntegerAttr(0, builtin.IndexType()))
        )
        one = rewriter.insert_op(
            arith.ConstantOp(builtin.IntegerAttr(1, builtin.IndexType()))
        )
        list_len = rewriter.insert_op(tensor.DimOp(op.li, zero))

        tensor_type = _list_type_to_tensor(op.li.type)

        rewriter_ip = rewriter.insertion_point

        for_body = Block([], arg_types=[builtin.IndexType()])
        ind_var = for_body.args[0]
        ind_var.name_hint = "_i"

        rewriter.insertion_point = InsertPoint.at_start(for_body)

        scalar = rewriter.insert_op(
            tensor.ExtractOp(op.li, [ind_var], tensor_type.element_type)
        )
        ListLangType.from_xdsl(scalar.result.type).print(rewriter, scalar.result)
        rewriter.insert_op(printf.PrintFormatOp(","))
        rewriter.insert_op(scf.YieldOp())

        rewriter.insertion_point = rewriter_ip

        rewriter.replace_matched_op(
            (
                printf.PrintFormatOp("["),
                scf.ForOp(
                    zero,
                    list_len,
                    one,
                    [],
                    for_body,
                ),
                printf.PrintFormatOp("]"),
            )
        )


class LowerRangeOp(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: list_dialect.RangeOp, rewriter: PatternRewriter):
        zero = rewriter.insert_op(
            arith.ConstantOp(builtin.IntegerAttr(0, builtin.IndexType()))
        )
        one = rewriter.insert_op(
            arith.ConstantOp(builtin.IntegerAttr(1, builtin.IndexType()))
        )

        tensor_length = rewriter.insert_op(arith.SubiOp(op.upper, op.lower))
        tensor_length_index = rewriter.insert_op(
            arith.IndexCastOp(tensor_length.result, builtin.IndexType())
        )
        tensor_length_index.result.name_hint = _name_hint_ext(
            op.result.name_hint, "length"
        )

        tensor_type = _list_type_to_tensor(op.result.type)
        start_tensor = rewriter.insert_op(
            tensor.EmptyOp((tensor_length_index.result,), tensor_type)
        )
        start_tensor.tensor.name_hint = _name_hint_ext(op.result.name_hint, "uninit")

        rewriter_ip = rewriter.insertion_point

        for_body = Block(
            [],
            arg_types=(builtin.IndexType(), tensor_type),
        )
        ind_var = for_body.args[0]
        ind_var.name_hint = "_i"
        tensor_arg = for_body.args[1]

        rewriter.insertion_point = InsertPoint.at_start(for_body)

        ind_var_32 = rewriter.insert_op(arith.IndexCastOp(ind_var, builtin.i32))
        offseted = rewriter.insert_op(arith.AddiOp(op.lower, ind_var_32))
        modified = rewriter.insert_op(
            tensor.InsertOp(offseted.result, tensor_arg, [ind_var])
        )
        modified.result.name_hint = _name_hint_ext(op.result.name_hint, "modified")
        rewriter.insert_op(scf.YieldOp(modified))

        rewriter.insertion_point = rewriter_ip

        for_op = scf.ForOp(
            zero,
            tensor_length_index,
            one,
            [start_tensor],
            for_body,
        )
        for_op.results[0].name_hint = op.result.name_hint
        rewriter.replace_matched_op(for_op)


class LowerListToTensor(ModulePass):
    """
    Lowers list dialect to a tensor-based representation.
    """

    name = "lower-list-to-tensor"

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    LowerLengthOp(),
                    LowerMapOp(),
                    LowerPrintOp(),
                    LowerRangeOp(),
                ]
            ),
        ).rewrite_module(op)


class WrapModuleInFunc(ModulePass):
    """
    Wraps the free-standing ops in a module into a func.func function.
    """

    name = "wrap-module-in-func"

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        old_module_region = op.detach_region(0)
        new_module_region = Region(
            [Block([func.FuncOp("main", ((), ()), old_module_region)])]
        )
        old_module_region.block.add_op(func.ReturnOp())
        op.add_region(new_module_region)



from xdsl.dialects.builtin import UnitAttr
from xdsl.dialects.x86.assembly import masked_source_str
from xdsl.dialects.x86.registers import AVX512MaskRegisterType, AVX512RegisterType
from xdsl.utils.test_value import create_ssa_value


def test_masked_source_str():
    src_reg_val = create_ssa_value(AVX512RegisterType.from_name("zmm0"))
    k_val = create_ssa_value(AVX512MaskRegisterType.from_name("k1"))

    assert masked_source_str(src_reg_val, k_val, None) == "zmm0 {k1}"
    assert masked_source_str(src_reg_val, k_val, UnitAttr()) == "zmm0 {k1}{z}"



from dataclasses import dataclass

from typing_extensions import override

from xdsl.backend.register_stack import RegisterStack
from xdsl.dialects.x86 import registers


@dataclass
class X86RegisterStack(RegisterStack):
    """
    Available x86-specific registers.
    """

    DEFAULT_ALLOCATABLE_REGISTERS = (
        *reversed(registers.GeneralRegisterType.allocatable_registers()),
        *reversed(registers.AVX2RegisterType.allocatable_registers()),
        *reversed(registers.AVX512RegisterType.allocatable_registers()),
    )

    @classmethod
    @override
    def default_allocatable_registers(cls):
        return cls.DEFAULT_ALLOCATABLE_REGISTERS



# RUN: python %s | filecheck %s

from ctypes import c_float, c_int32, c_int64, c_size_t

from xdsl.dialects import builtin
from xdsl.frontend.pyast.context import PyASTContext

ctx = PyASTContext(post_transforms=[])
ctx.register_type(c_int32, builtin.i32)
ctx.register_type(c_int64, builtin.i64)
ctx.register_type(c_size_t, builtin.IndexType())
ctx.register_type(bool, builtin.i1)
ctx.register_type(c_float, builtin.f32)
ctx.register_type(float, builtin.f64)


# CHECK: @boolean(%{{.*}} : i1)
@ctx.parse_program
def boolean(x: bool):
    return


print(boolean.module)


# CHECK: @signless(%{{.*}} : i32, %{{.*}} : i64)
@ctx.parse_program
def signless(x: c_int32, y: c_int64):
    return


print(signless.module)


# CHECK: @indexed(%{{.*}} : index)
@ctx.parse_program
def indexed(x: c_size_t):
    return


print(indexed.module)


# CHECK: @fp(%{{.*}} : f32, %{{.*}} : f64)
@ctx.parse_program
def fp(x: c_float, y: float):
    return


print(fp.module)



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterator, Sequence
from dataclasses import dataclass
from enum import Enum, auto
from itertools import chain
from typing import TYPE_CHECKING

from typing_extensions import assert_never

# Used for cyclic dependencies in type hints
if TYPE_CHECKING:
    from xdsl.ir.affine import AffineMap


@dataclass(frozen=True)
class AffineExpr(ABC):
    """
    An AffineExpr models an affine expression, which is a linear combination of
    dimensions with integer coefficients. For example, 2 * d0 + 3 * d1 is an
    affine expression, where d0, d1 are dimensions. An AffineExpr can be
    parameterized by symbols. AffineExpr also allows further extensions of an
    affine expression. Quasi-affine expressions, i.e. Integer division and
    modulo with a constant are allowed. For example, 2 * d0 + 3 * d1 + 4
    floordiv 5 is a quasi-affine expression. Semi-affine expressions i.e.
    Integer division and modulo with a symbol are also allowed. For example, 2
    * d0 + 3 * d1 + 4 floordiv s0 is a semi-affine expression.
    """

    @staticmethod
    def constant(value: int) -> AffineExpr:
        return AffineConstantExpr(value)

    @staticmethod
    def dimension(position: int) -> AffineExpr:
        return AffineDimExpr(position)

    @staticmethod
    def symbol(position: int) -> AffineExpr:
        return AffineSymExpr(position)

    @staticmethod
    def binary(
        kind: AffineBinaryOpKind,
        lhs: AffineExpr,
        rhs: AffineExpr,
    ) -> AffineExpr:
        """
        Builds a binary expression of the given kind using the operator function associated with that kind.
        As a consequence, binary expressions are simplified during construction.
        This may lead to the resulting expression not being of the type `AffineBinaryOpExpr`, but of the type that binary op is
        simplified to.
        This simplification does not occur when an `AffineBinaryOpExpr` is directly created using its constructor.

        Example:
        An expression of kind `AffineBinaryKind.Add` is built using the `AffineExpr.__add__` function.
        If both `rhs` and `lhs` are `AffineConstantExprs` this function returns an `AffineConstantExpr` of value `rhs` + `lhs`.
        """

        match kind:
            case AffineBinaryOpKind.Add:
                return lhs + rhs
            case AffineBinaryOpKind.Mul:
                return lhs * rhs
            case AffineBinaryOpKind.Mod:
                return lhs % rhs
            case AffineBinaryOpKind.FloorDiv:
                return lhs // rhs
            case AffineBinaryOpKind.CeilDiv:
                return lhs.ceil_div(rhs)
            case _:
                assert_never(kind)

    @staticmethod
    def from_flat_form(
        flat_exprs: Sequence[int],
        num_dims: int,
        num_symbols: int,
        local_exprs: Sequence[AffineExpr],
    ) -> AffineExpr:
        """
        Constructs an affine expression from a flat list of coefficients.
        If there are local identifiers (neither dimensional nor symbolic) that appear in
        the sum of products expression, `local_exprs` is expected to have the AffineExpr
        for it, and is substituted into.
        The list `flat_exprs` is expected to be in the format [*dims, *symbols, *locals,
        constant term].
        """
        assert len(flat_exprs) == num_dims + num_symbols + len(local_exprs) + 1, (
            f"unexpected number of local expressions {len(local_exprs)}, expected "
            f"{len(flat_exprs) - num_dims - num_symbols - 1}"
        )

        expr = sum(
            (
                e * f
                for e, f in zip(
                    chain(
                        (AffineExpr.dimension(d) for d in range(num_dims)),
                        (AffineExpr.symbol(s) for s in range(num_symbols)),
                        local_exprs,
                    ),
                    flat_exprs[:-1],
                    strict=True,
                )
                if f != 0
            ),
            start=AffineExpr.constant(0),
        )

        # Constant term
        const_term = flat_exprs[-1]
        if const_term != 0:
            expr = expr + const_term

        return expr

    def simplify(self, num_dims: int, num_symbols: int) -> AffineExpr:
        """
        Simplify the affine expression by flattening it and reconstructing it.
        """
        if not self.is_pure_affine():
            # Simplify semi-affine expressions separately
            raise NotImplementedError(
                "Simplification of semi-affine expressions is not implemented yet."
            )

        flattener = SimpleAffineExprFlattener(num_dims, num_symbols)
        return flattener.simplify(self)

    def compose(self, map: AffineMap) -> AffineExpr:
        """
        Compose with an AffineMap.

        Returns the composition of this AffineExpr with map.

        Prerequisites: this and map are composable, i.e. that the number of
        AffineDimExpr of this is smaller than the number of results of map.
        If a result of a map does not have a corresponding AffineDimExpr, that result
        simply does not appear in the produced AffineExpr.

        Example:
        ```
        expr: d0 + d2
        map: (d0, d1, d2)[s0, s1] -> (d0 + s1, d1 + s0, d0 + d1 + d2)
        returned expr: d0 * 2 + d1 + d2 + s1
        ```
        """
        return self.replace_dims_and_symbols(map.results, ())

    def replace_dims_and_symbols(
        self, new_dims: Sequence[AffineExpr], new_symbols: Sequence[AffineExpr]
    ) -> AffineExpr:
        """Replace the symbols and indices in this map with the ones provided."""
        if isinstance(self, AffineConstantExpr):
            return self

        if isinstance(self, AffineDimExpr):
            if self.position >= len(new_dims):
                return self
            return new_dims[self.position]

        if isinstance(self, AffineSymExpr):
            if self.position >= len(new_symbols):
                return self
            return new_symbols[self.position]

        if isinstance(self, AffineBinaryOpExpr):
            lhs = self.lhs.replace_dims_and_symbols(new_dims, new_symbols)
            rhs = self.rhs.replace_dims_and_symbols(new_dims, new_symbols)

            return AffineExpr.binary(
                lhs=lhs,
                rhs=rhs,
                kind=self.kind,
            )

        raise ValueError("Unreachable")

    def eval(self, dims: Sequence[int], symbols: Sequence[int]) -> int:
        """Evaluate the affine expression with the given dimension and symbol values."""
        if isinstance(self, AffineConstantExpr):
            return self.value

        if isinstance(self, AffineDimExpr):
            return dims[self.position]
        if isinstance(self, AffineSymExpr):
            return symbols[self.position]

        if isinstance(self, AffineBinaryOpExpr):
            lhs = self.lhs.eval(dims, symbols)
            rhs = self.rhs.eval(dims, symbols)

            if self.kind == AffineBinaryOpKind.Add:
                return lhs + rhs
            elif self.kind == AffineBinaryOpKind.Mul:
                return lhs * rhs
            elif self.kind == AffineBinaryOpKind.Mod:
                return lhs % rhs
            elif self.kind == AffineBinaryOpKind.FloorDiv:
                return lhs // rhs
            elif self.kind == AffineBinaryOpKind.CeilDiv:
                return -(-lhs // rhs)

        raise ValueError("Unreachable")

    def _try_fold_constant(
        self, other: AffineExpr, kind: AffineBinaryOpKind
    ) -> AffineExpr | None:
        if not isinstance(self, AffineConstantExpr):
            return None
        if not isinstance(other, AffineConstantExpr):
            return None

        match kind:
            case AffineBinaryOpKind.Add:
                return AffineExpr.constant(self.value + other.value)
            case AffineBinaryOpKind.Mul:
                return AffineExpr.constant(self.value * other.value)
            case AffineBinaryOpKind.Mod:
                return AffineExpr.constant(self.value % other.value)
            case AffineBinaryOpKind.FloorDiv:
                return AffineExpr.constant(self.value // other.value)
            case AffineBinaryOpKind.CeilDiv:
                return AffineExpr.constant(-(-self.value // other.value))

    def _simplify_add(self, other: AffineExpr) -> AffineExpr | None:
        """Simplify addition. Constant is assumed to be on RHS."""
        # Fold constants.
        if fold := self._try_fold_constant(other, AffineBinaryOpKind.Add):
            return fold
        # Ignore addition with 0.
        if isinstance(other, AffineConstantExpr) and other.value == 0:
            return self
        # Fold (expr + constant) + constant.
        if isinstance(self, AffineBinaryOpExpr) and self.kind == AffineBinaryOpKind.Add:
            if fold := self.rhs._try_fold_constant(other, AffineBinaryOpKind.Add):
                return self.lhs + fold
        return None

    def __add__(self, other: AffineExpr | int) -> AffineExpr:
        if isinstance(other, int):
            other = AffineExpr.constant(other)
        # Canonicalize the expression so that the constant is always on the RHS.
        if isinstance(self, AffineConstantExpr):
            self, other = other, self
        if simplified := self._simplify_add(other):
            return simplified
        return AffineBinaryOpExpr(AffineBinaryOpKind.Add, self, other)

    def __radd__(self, other: AffineExpr | int) -> AffineExpr:
        return self.__add__(other)

    def __neg__(self) -> AffineExpr:
        if isinstance(self, AffineConstantExpr):
            return AffineExpr.constant(-self.value)
        return self * -1

    def __sub__(self, other: AffineExpr | int) -> AffineExpr:
        return self + (-1 * other)

    def __rsub__(self, other: AffineExpr | int) -> AffineExpr:
        return self.__sub__(other)

    def _simplify_mul(self, other: AffineExpr) -> AffineExpr | None:
        """Simplify multiplication. Constant is assumed to be on RHS."""
        # Fold constant.
        if fold := self._try_fold_constant(other, AffineBinaryOpKind.Mul):
            return fold
        # Ignore multiplication by 1.
        if isinstance(other, AffineConstantExpr) and other.value == 1:
            return self
        # Fold (expr * constant) * constant.
        if isinstance(self, AffineBinaryOpExpr) and self.kind == AffineBinaryOpKind.Mul:
            if fold := self.rhs._try_fold_constant(other, AffineBinaryOpKind.Mul):
                return self.lhs * fold
        # Fold (expr + expr) * constant.
        if (
            isinstance(self, AffineBinaryOpExpr)
            and self.kind == AffineBinaryOpKind.Add
            and isinstance(other, AffineConstantExpr)
        ):
            return self.lhs * other + self.rhs * other
        return None

    def __mul__(self, other: AffineExpr | int) -> AffineExpr:
        if isinstance(other, int):
            other = AffineExpr.constant(other)
        # Canonicalize the expression so that the constant is always on the RHS.
        if isinstance(self, AffineConstantExpr):
            self, other = other, self
        if simplified := self._simplify_mul(other):
            return simplified
        if not isinstance(other, AffineConstantExpr):
            # TODO (#1087): MLIR also supports multiplication by symbols also, making
            # maps semi-affine. Currently, we do not implement semi-affine maps.
            raise NotImplementedError(
                f"Multiplication with non-constant (semi-affine) is not supported yet self: {self} other: {other}"
            )
        return AffineBinaryOpExpr(AffineBinaryOpKind.Mul, self, other)

    def __rmul__(self, other: AffineExpr | int) -> AffineExpr:
        return self.__mul__(other)

    def __floordiv__(self, other: AffineExpr | int) -> AffineExpr:
        if isinstance(other, int):
            other = AffineExpr.constant(other)

        # Fold constants.
        if fold := self._try_fold_constant(other, AffineBinaryOpKind.FloorDiv):
            return fold

        if not isinstance(other, AffineConstantExpr):
            # TODO (#1087): MLIR also supports floor-division by symbols also, making
            # maps semi-affine. Currently, we do not implement semi-affine maps.
            raise NotImplementedError(
                "Floor division with non-constant (semi-affine) is not supported yet"
            )
        # TODO (#1086): Simplify floor division here before returning.
        return AffineBinaryOpExpr(AffineBinaryOpKind.FloorDiv, self, other)

    def ceil_div(self, other: AffineExpr | int) -> AffineExpr:
        if isinstance(other, int):
            other = AffineExpr.constant(other)

        # Fold constants.
        if fold := self._try_fold_constant(other, AffineBinaryOpKind.CeilDiv):
            return fold

        if not isinstance(other, AffineConstantExpr):
            # TODO (#1087): MLIR also supports ceil-division by symbols also, making
            # maps semi-affine. Currently, we do not implement semi-affine maps.
            raise NotImplementedError(
                "Ceil division with non-constant (semi-affine) is not supported yet"
            )
        # TODO (#1086): Simplify ceil division here before returning.
        return AffineBinaryOpExpr(AffineBinaryOpKind.CeilDiv, self, other)

    def __mod__(self, other: AffineExpr | int) -> AffineExpr:
        if isinstance(other, int):
            other = AffineExpr.constant(other)

        # Fold constants.
        if fold := self._try_fold_constant(other, AffineBinaryOpKind.Mod):
            return fold

        if not isinstance(other, AffineConstantExpr):
            # TODO (#1087): MLIR also supports Mod by symbols also, making maps
            # semi-affine. Currently, we do not implement semi-affine maps.
            raise NotImplementedError(
                "Mod with non-constant (semi-affine) is not supported yet"
            )
        # TODO (#1086): Simplify modulo here before returning.
        return AffineBinaryOpExpr(AffineBinaryOpKind.Mod, self, other)

    def dfs(self) -> Iterator[AffineExpr]:
        """
        Iterates nodes in depth-first order (parent-left-right).

        See external [documentation](https://en.wikipedia.org/wiki/Depth-first_search).
        """
        yield self

    def post_order(self) -> Iterator[AffineExpr]:
        """
        Iterates nodes in pre-order (left-right-parent).

        See external [documentation](https://en.wikipedia.org/wiki/Tree_traversal).
        """
        yield self

    def used_dims(self) -> set[int]:
        return {expr.position for expr in self.dfs() if isinstance(expr, AffineDimExpr)}

    @abstractmethod
    def is_pure_affine(self) -> bool:
        """
        Returns true if this is a pure affine expression, i.e., multiplication,
        floordiv, ceildiv, and mod is only allowed w.r.t constants.
        """
        ...


class AffineBinaryOpKind(Enum):
    """Enum for the kind of storage node used in AffineExpr."""

    Add = auto()
    Mul = auto()
    Mod = auto()
    FloorDiv = auto()
    CeilDiv = auto()

    def get_token(self) -> str:
        match self:
            case AffineBinaryOpKind.Add:
                return "+"
            case AffineBinaryOpKind.Mul:
                return "*"
            case AffineBinaryOpKind.Mod:
                return "mod"
            case AffineBinaryOpKind.FloorDiv:
                return "floordiv"
            case AffineBinaryOpKind.CeilDiv:
                return "ceildiv"


@dataclass(frozen=True)
class AffineBinaryOpExpr(AffineExpr):
    """An affine expression storage node representing a binary operation."""

    kind: AffineBinaryOpKind
    lhs: AffineExpr
    rhs: AffineExpr

    def dfs(self) -> Iterator[AffineExpr]:
        yield self
        yield from self.lhs.dfs()
        yield from self.rhs.dfs()

    def post_order(self) -> Iterator[AffineExpr]:
        yield from self.lhs.post_order()
        yield from self.rhs.post_order()
        yield self

    def is_pure_affine(self) -> bool:
        match self.kind:
            case AffineBinaryOpKind.Add:
                return self.lhs.is_pure_affine() and self.rhs.is_pure_affine()
            case AffineBinaryOpKind.Mul:
                # Multiplication is only allowed with a constant on the right
                # or left for pure affine expressions.
                # Check if either lhs or rhs is a constant and the other is pure affine.
                lhs_is_const = isinstance(self.lhs, AffineConstantExpr)
                rhs_is_const = isinstance(self.rhs, AffineConstantExpr)

                return (
                    (lhs_is_const or rhs_is_const)
                    and self.lhs.is_pure_affine()
                    and self.rhs.is_pure_affine()
                )
            case (
                AffineBinaryOpKind.Mod
                | AffineBinaryOpKind.FloorDiv
                | AffineBinaryOpKind.CeilDiv
            ):
                # Mod, floordiv, ceildiv are only allowed with a constant on the right for pure affine
                return (
                    isinstance(self.rhs, AffineConstantExpr)
                    and self.lhs.is_pure_affine()
                )

    def __str__(self) -> str:
        return f"({self.lhs} {self.kind.get_token()} {self.rhs})"


@dataclass(frozen=True)
class AffineDimExpr(AffineExpr):
    """An affine expression storage node representing a dimension."""

    position: int

    def is_pure_affine(self) -> bool:
        return True

    def __str__(self) -> str:
        return f"d{self.position}"


@dataclass(frozen=True)
class AffineSymExpr(AffineExpr):
    """An affine expression storage node representing a symbol."""

    position: int

    def is_pure_affine(self) -> bool:
        return True

    def __str__(self) -> str:
        return f"s{self.position}"


@dataclass(frozen=True)
class AffineConstantExpr(AffineExpr):
    """An affine expression storage node representing a constant."""

    value: int

    def is_pure_affine(self) -> bool:
        return True

    def __str__(self) -> str:
        return f"{self.value}"


class SimpleAffineExprFlattener:
    """
    This class is used to flatten a pure affine expression (AffineExpr, which is in a
    tree form) into a sum of products (with respect to constants) when possible, thereby
    simplifying the expression. For modulo, floordiv, or ceildiv expressions, an
    additional identifier, called a local identifier, is introduced to rewrite the
    expression as a sum of product affine expression. Each local identifier is always,
    by construction, a floordiv of a pure add/mul affine function of dimensional,
    symbolic, and other local identifiers, in a non-mutually recursive way. Thus, every
    local identifier can ultimately always be recovered as an affine function of
    dimensional and symbolic identifiers (involving floordiv's); note, however, that by
    AffineExpr construction, some floordiv combinations are converted to mod's.
    The result of the flattening is a flattened expression and a set of
    constraints involving just the local variables.

    For example, `d2 + (d0 + d1) // 4` is flattened to `d2 + q` where `q` is
    the local variable introduced, with `localVarCst` containing
    `4*q <= d0 + d1 <= 4*q + 3`.

    The simplification performed includes the accumulation of contributions for
    each dimensional and symbolic identifier together, the simplification of
    floordiv/ceildiv/mod expressions, and other simplifications that in turn
    happen as a result. A simplification that this flattening naturally performs
    is simplifying the numerator and denominator of floordiv/ceildiv, and
    folding a modulo expression to zero, if possible. Three examples are below:

    ```
    (d0 + 3 * d1) + d0) - 2 * d1) - d0    simplified to     d0 + d1
    (d0 - d0 % 4 + 4) % 4                 simplified to     0
    (3*d0 + 2*d1 + d0) // 2 + d1          simplified to     2*d0 + 2*d1
    ```

    The way the flattening works for the second example is as follows: `d0 % 4` is
    replaced by `d0 - 4*q` with `q` being introduced; the expression then simplifies
    to: `(d0 - (d0 - 4q) + 4) = 4q + 4`, modulo of which with respect to 4
    simplifies to zero. Note that an affine expression may not always be
    expressible purely as a sum of products involving just the original
    dimensional and symbolic identifiers due to the presence of
    modulo/floordiv/ceildiv expressions that may not be eliminated after
    simplification; in such cases, the final expression can be reconstructed by
    replacing the local identifiers with their corresponding explicit form
    stored in `localExprs` (note that each of the explicit forms itself would
    have been simplified).

    The expression walk method here performs a linear time post-order walk that
    performs the above simplifications through visit methods, with partial
    results being stored in `operandExprStack`. When a parent expr is visited,
    the flattened expressions corresponding to its two operands would already be
    on the stack—the parent expression looks at the two flattened expressions
    and combines the two. It pops off the operand expressions and pushes the
    combined result (although this is done in-place on its LHS operand expr).
    When the walk is completed, the flattened form of the top-level expression
    would be left on the stack.

    A flattener can be repeatedly used for multiple affine expressions that bind
    to the same operands, for example, for all result expressions of an
    AffineMap or AffineValueMap. In such cases, using it for multiple
    expressions is more efficient than creating a new flattener for each
    expression since common identical div and mod expressions appearing across
    different expressions are mapped to the same local identifier (same column
    position in `localVarCst`).
    """

    # Flattend expression layout: [dims, symbols, locals, constant]
    # Stack that holds the LHS and RHS operands while visiting a binary op expr.
    operand_expr_stack: list[list[int]]
    """
    Flattend expression layout: [dims, symbols, locals, constant]
    Stack that holds the LHS and RHS operands while visiting a binary op expr.
    """
    num_dims: int
    num_symbols: int
    local_exprs: list[AffineExpr]

    def __init__(self, num_dims: int, num_symbols: int) -> None:
        self.operand_expr_stack = []
        self.num_dims = num_dims
        self.num_symbols = num_symbols
        self.local_exprs = []

    def visit_mul_expr(self, expr: AffineBinaryOpExpr) -> None:
        """
        In pure affine t = expr * c, we multiply each coefficient of lhs with c.
        In case of semi affine multiplication expressions, `t = expr * symbolic_expr`,
        introduce a local variable `p (= expr * symbolic_expr)`, and the affine expression
        `expr * symbolic_expr`` is added to `localExprs`.
        """
        assert len(self.operand_expr_stack) >= 2
        rhs = self.operand_expr_stack.pop()
        lhs = self.operand_expr_stack.pop()

        if not isinstance(expr.rhs, AffineConstantExpr):
            # Flatten semi-affine multiplication expressions by introducing a local
            # variable in place of the product; the affine expression
            # corresponding to the quantifier is added to `localExprs`.
            raise NotImplementedError("Semi-affine map flattening not implemented")

        rhs_const = rhs[self.get_constant_index()]
        self.operand_expr_stack.append([l * rhs_const for l in lhs])

    def visit_add_expr(self, expr: AffineBinaryOpExpr) -> None:
        assert len(self.operand_expr_stack) >= 2
        rhs = self.operand_expr_stack.pop()
        lhs = self.operand_expr_stack.pop()
        assert len(lhs) == len(rhs)
        self.operand_expr_stack.append([l + r for l, r in zip(lhs, rhs, strict=True)])

    def visit_dim_expr(self, expr: AffineDimExpr) -> None:
        row = [0] * self.get_num_cols()
        assert expr.position < self.num_dims, "Inconsistent number of dims"
        row[self.get_dim_start_index() + expr.position] = 1
        self.operand_expr_stack.append(row)

    def visit_symbol_expr(self, expr: AffineSymExpr) -> None:
        # Equivalent to SimpleAffineExprFlattener::visitSymbolExpr
        row = [0] * self.get_num_cols()
        assert expr.position < self.num_symbols, "Inconsistent number of symbols"
        row[self.get_symbol_start_index() + expr.position] = 1
        self.operand_expr_stack.append(row)

    def visit_constant_expr(self, expr: AffineConstantExpr) -> None:
        # Equivalent to SimpleAffineExprFlattener::visitConstantExpr
        row = [0] * self.get_num_cols()
        row[self.get_constant_index()] = expr.value
        self.operand_expr_stack.append(row)

    def visit_div_expr(self, expr: AffineBinaryOpExpr, *, is_ceil: bool) -> None:
        raise NotImplementedError("Simplifying div expr not yet implemented")

    def visit_mod_expr(self, expr: AffineBinaryOpExpr) -> None:
        raise NotImplementedError("Simplifying mod expr not yet implemented")

    def simplify(self, expr: AffineExpr):
        for inner in expr.post_order():
            match inner:
                case AffineBinaryOpExpr():
                    match inner.kind:
                        case AffineBinaryOpKind.Mul:
                            self.visit_mul_expr(inner)
                        case AffineBinaryOpKind.Add:
                            self.visit_add_expr(inner)
                        case AffineBinaryOpKind.Mod:
                            self.visit_mod_expr(inner)
                        case AffineBinaryOpKind.FloorDiv:
                            self.visit_div_expr(inner, is_ceil=False)
                        case AffineBinaryOpKind.CeilDiv:
                            self.visit_div_expr(inner, is_ceil=True)
                case AffineDimExpr():
                    self.visit_dim_expr(inner)
                case AffineConstantExpr():
                    self.visit_constant_expr(inner)
                case AffineSymExpr():
                    self.visit_symbol_expr(inner)
                case _:
                    raise ValueError("Unreachable")

        return AffineExpr.from_flat_form(
            self.operand_expr_stack.pop(),
            self.num_dims,
            self.num_symbols,
            self.local_exprs,
        )

    def get_num_cols(self) -> int:
        return self.num_dims + self.num_symbols + len(self.local_exprs) + 1

    def get_constant_index(self) -> int:
        return self.get_num_cols() - 1

    def get_local_var_start_index(self) -> int:
        return self.num_dims + self.num_symbols

    def get_symbol_start_index(self) -> int:
        return self.num_dims

    def get_dim_start_index(self) -> int:
        return 0



import pytest

from xdsl.backend.block_throughput_cost_model import MCABlockThroughputCostModel
from xdsl.builder import Builder
from xdsl.dialects import x86_func

llvm_mca_available = MCABlockThroughputCostModel.is_installed()


@pytest.mark.skipif(not llvm_mca_available, reason="llvm-mca is not installed")
def test_mca_reporter_x86():
    @Builder.implicit_region
    def trivial_x86_func():
        x86_func.RetOp()

    reporter = MCABlockThroughputCostModel(
        target="x86_64-unknown-linux-gnu", arch="skylake"
    )
    estimated_cost = reporter.estimate_throughput(trivial_x86_func.block)
    assert estimated_cost is not None, (
        "MCA reporter should return a valid cost estimate"
    )



import pytest

from xdsl.dialects import arith, eqsat, test
from xdsl.dialects.builtin import IntegerAttr, i32
from xdsl.traits import ConstantLike
from xdsl.utils.exceptions import DiagnosticException


def test_const_eclass_construction():
    constant_op = arith.ConstantOp(IntegerAttr.from_int_and_width(42, 64))

    const_eclass = eqsat.ConstantEClassOp(constant_op.result)
    trait = const_eclass.get_trait(ConstantLike)
    assert trait is not None
    assert trait.get_constant_value(const_eclass) == IntegerAttr.from_int_and_width(
        42, 64
    )

    non_constant_op = test.TestOp(result_types=(i32,))
    with pytest.raises(
        DiagnosticException,
        match="The argument of a ConstantEClass must be a constant-like operation.",
    ):
        eqsat.ConstantEClassOp(non_constant_op.results[0])



import pytest

from xdsl.dialects import complex
from xdsl.dialects.builtin import (
    AnyFloat,
    ArrayAttr,
    ComplexType,
    FloatAttr,
    IntAttr,
    IntegerAttr,
    IntegerType,
    f16,
    f32,
    f64,
    i1,
    i8,
    i16,
    i32,
)
from xdsl.traits import ConstantLike


def test_constant_construction():
    c1 = complex.ConstantOp(
        value=ArrayAttr([IntAttr(42), IntAttr(43)]),
        result_type=complex.ComplexType(i32),
    )
    constantlike = c1.get_trait(ConstantLike)
    assert constantlike is not None
    assert constantlike.get_constant_value(c1) == ArrayAttr([IntAttr(42), IntAttr(43)])


class Test_constant_op_helper_constructors:
    @pytest.mark.parametrize(
        "real, imag, type",
        [
            (0, 1, i1),
            (1, 2, i8),
            (24, 19, i16),
            (33, 88, i32),
        ],
    )
    def test_from_ints(self, real: int, imag: int, type: IntegerType):
        cst = complex.ConstantOp.from_ints((real, imag), type)
        assert cst.value.data[0].type == type
        assert cst.value == ArrayAttr(
            [IntegerAttr(real, type), IntegerAttr(imag, type)]
        )
        assert cst.result_types[0] == ComplexType(type)

    @pytest.mark.parametrize(
        "real, imag, type",
        [
            (0, 1, f16),
            (-1, 2.2, f32),
            (24.45, 19.0, f64),
        ],
    )
    def test_from_floats(self, real: float, imag: float, type: AnyFloat):
        cst = complex.ConstantOp.from_floats((real, imag), type)
        assert cst.value.data[0].type == type
        assert cst.value == ArrayAttr([FloatAttr(real, type), FloatAttr(imag, type)])
        assert cst.result_types[0] == ComplexType(type)



from dataclasses import dataclass

from xdsl.builder import ImplicitBuilder
from xdsl.context import Context
from xdsl.dialects import arith, builtin, linalg, memref, vector
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.utils.exceptions import DiagnosticException
from xdsl.utils.hints import isa

_index_type = builtin.IndexType()


@dataclass
class VectorizeMatmulOp(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: linalg.MatmulOp, rewriter: PatternRewriter, /):
        # C += A * B
        # C: M x N, A: M x K, B: K x N

        a, b = op.inputs
        c = op.outputs[0]

        a_type = a.type
        b_type = b.type
        c_type = c.type

        # Only handle matmul on memrefs for now
        if (
            not isa(a_type, builtin.MemRefType)
            or not isa(b_type, builtin.MemRefType)
            or not isa(c_type, builtin.MemRefType)
        ):
            raise DiagnosticException(
                "Vectorizing matmul on tensors not yet implemented."
            )

        M, K = a_type.get_shape()
        _K, N = b_type.get_shape()
        _M, _N = c_type.get_shape()

        assert M == _M
        assert N == _N
        assert K == _K
        assert M != -1
        assert N != -1
        assert K != -1

        vector_type = builtin.VectorType(a_type.element_type, (N,))

        # All operations created inside this block are inserted before the matched op
        with ImplicitBuilder(rewriter):
            # Insert all the integer constants we'll need to index into the matrices
            constants = tuple(
                arith.ConstantOp(builtin.IntegerAttr(i, _index_type)).result
                for i in range(max(M, N, K))
            )
            # Zero for convenience
            c0 = constants[0]

            # Load the rows of C as vectors
            c_rows = [
                vector.LoadOp(c, (constants[m], c0), vector_type).result
                for m in range(M)
            ]

            # Load the rows of B as vectors
            b_rows = tuple(
                vector.LoadOp(b, (constants[k], c0), vector_type).result
                for k in range(K)
            )

            for m in range(M):
                # Load the mth column of A as scalars
                a_col = tuple(
                    memref.LoadOp.get(a, (constants[m], constants[k])).res
                    for k in range(K)
                )
                # Broadcast the mth column of A to vectors
                a_col_vectors = tuple(
                    vector.BroadcastOp(a_col[k], vector_type) for k in range(K)
                )

                for k in range(K):
                    # Accumulate the dot product of rows of B with A's column
                    # The list c_rows is updated in place for convenience, but we're
                    # really creating a new SSA value on each iteration
                    c_rows[m] = vector.FMAOp(a_col_vectors[k], b_rows[k], c_rows[m]).res

            for m in range(M):
                vector.StoreOp(c_rows[m], c, (constants[m], c0))

        rewriter.erase_matched_op()


@dataclass(frozen=True)
class TestVectorizeMatmulPass(ModulePass):
    """
    A test pass vectorizing linalg.matmul with a specific vectorization strategy.
    """

    name = "test-vectorize-matmul"

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        PatternRewriteWalker(
            VectorizeMatmulOp(), apply_recursively=False
        ).rewrite_module(op)



from dataclasses import dataclass

from ordered_set import OrderedSet

from xdsl.builder import Builder
from xdsl.context import Context
from xdsl.dialects import builtin, x86, x86_func
from xdsl.dialects.x86.registers import (
    R12,
    R13,
    R14,
    R15,
    RBP,
    RBX,
    RSP,
    GeneralRegisterType,
)
from xdsl.passes import ModulePass
from xdsl.rewriter import InsertPoint

X86_CALLEE_SAVED_REGISTERS = [RBX, RBP, R12, R13, R14, R15]
"""
Registers that should be the same after the called function returns to the caller, see
[external documentation](https://refspecs.linuxbase.org/elf/x86_64-abi-0.21.pdf).
"""


@dataclass(frozen=True)
class X86PrologueEpilogueInsertion(ModulePass):
    name = "x86-prologue-epilogue-insertion"

    def _process_function(self, func: x86_func.FuncOp) -> None:
        used_callee_preserved_registers = OrderedSet(
            res.type
            for op in func.walk()
            if not isinstance(op, x86.GetRegisterOp)
            for res in op.results
            if isinstance(res.type, GeneralRegisterType)
            if res.type in X86_CALLEE_SAVED_REGISTERS
        )

        if not used_callee_preserved_registers:
            return

        builder = Builder(InsertPoint.at_start(func.body.blocks[0]))
        sp_register = builder.insert(x86.GetRegisterOp(RSP))

        # Build the prologue at the beginning of the function.
        for reg in used_callee_preserved_registers:
            reg_op = builder.insert(x86.GetRegisterOp(reg))
            builder.insert(x86.S_PushOp(rsp_in=sp_register, source=reg_op))

        # Now build the epilogue right before every return operation.
        for block in func.body.blocks:
            ret_op = block.last_op
            if not isinstance(ret_op, x86_func.RetOp):
                continue
            builder = Builder(InsertPoint.before(ret_op))
            for reg in reversed(used_callee_preserved_registers):
                builder.insert(x86.D_PopOp(rsp_in=sp_register, destination=reg))

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        for func in op.walk():
            if not isinstance(func, x86_func.FuncOp):
                continue

            if not func.body.blocks:
                continue

            self._process_function(func)



"""
This file implements some of the core data structures used in the pdl-to-pdl-interp conversion.
"""

from abc import ABC
from dataclasses import dataclass
from typing import Optional

from xdsl.dialects import pdl
from xdsl.dialects.builtin import (
    ArrayAttr,
    TypeAttribute,
)
from xdsl.ir import (
    Attribute,
)


@dataclass(frozen=True)
class Position(ABC):
    """The position class encodes a location in a pattern.
    Each pattern has a root position. From there, other positions can be reached representing operands, results, and more.
    """

    parent: Optional["Position"] = None

    def get_operation_depth(self) -> int:
        """Returns depth of first ancestor operation position"""
        op = self.get_base_operation()
        return op.depth

    def get_base_operation(self) -> "OperationPosition":
        pos = self
        while not isinstance(pos, OperationPosition):
            assert pos.parent is not None
            pos = pos.parent
        return pos


@dataclass(frozen=True, kw_only=True)
class OperationPosition(Position):
    """Represents an operation in the IR"""

    depth: int

    def is_root(self) -> bool:
        return self.depth == 0

    def is_operand_defining_op(self) -> bool:
        return isinstance(self.parent, OperandPosition | OperandGroupPosition)

    def __repr__(self):
        if self.is_root():
            return "root"
        else:
            return self.parent.__repr__() + ".defining_op"

    def get_operand(self, operand_num: int) -> "OperandPosition":
        return OperandPosition(operand_number=operand_num, parent=self)

    def get_operand_group(
        self, group_num: int, is_variadic: bool
    ) -> "OperandGroupPosition":
        return OperandGroupPosition(
            group_number=group_num, is_variadic=is_variadic, parent=self
        )

    def get_all_operands(self) -> "OperandGroupPosition":
        return OperandGroupPosition(group_number=None, is_variadic=True, parent=self)

    def get_result(self, result_num: int) -> "ResultPosition":
        return ResultPosition(result_number=result_num, parent=self)

    def get_result_group(
        self, group_num: int | None, is_variadic: bool
    ) -> "ResultGroupPosition":
        return ResultGroupPosition(
            group_number=group_num, is_variadic=is_variadic, parent=self
        )

    def get_all_results(self) -> "ResultGroupPosition":
        return ResultGroupPosition(group_number=None, is_variadic=True, parent=self)

    def get_attribute(self, attr_name: str) -> "AttributePosition":
        return AttributePosition(attribute_name=attr_name, parent=self)

    def get_attribute_literal(self, value: Attribute) -> "AttributeLiteralPosition":
        return AttributeLiteralPosition(value=value, parent=None)


class ValuePosition(Position):
    def get_defining_op(self) -> OperationPosition:
        return OperationPosition(depth=self.get_operation_depth() + 1, parent=self)

    def get_type(self) -> "TypePosition":
        return TypePosition(parent=self)

    def get_type_literal(self, value: Attribute) -> "TypeLiteralPosition":
        return TypeLiteralPosition(value=value, parent=None)

    def get_users(self, use_representative: bool) -> "UsersPosition":
        return UsersPosition(parent=self, use_representative=use_representative)


@dataclass(frozen=True, kw_only=True)
class OperandPosition(ValuePosition):
    """Represents an operand of an operation"""

    operand_number: int

    def __repr__(self):
        return f"{self.parent.__repr__()}.operand[{self.operand_number}]"


@dataclass(frozen=True, kw_only=True)
class OperandGroupPosition(Position):
    """Represents a group of operands"""

    group_number: int | None
    is_variadic: bool

    def get_defining_op(self) -> "OperationPosition":
        return OperationPosition(depth=self.get_operation_depth() + 1, parent=self)

    def get_type(self) -> "TypePosition":
        return TypePosition(parent=self)


@dataclass(frozen=True, kw_only=True)
class ResultPosition(ValuePosition):
    """Represents a result of an operation"""

    result_number: int

    def __repr__(self):
        return f"{self.parent.__repr__()}.result[{self.result_number}]"


@dataclass(frozen=True, kw_only=True)
class AttributePosition(Position):
    """Represents an attribute of an operation"""

    attribute_name: str

    def __repr__(self):
        return f"{self.parent.__repr__()}.attribute[{self.attribute_name}]"

    def get_type(self) -> "TypePosition":
        return TypePosition(parent=self)


@dataclass(frozen=True)
class TypePosition(Position):
    """Represents the type of a value"""

    def __repr__(self):
        return f"{self.parent.__repr__()}.type"

    def get_type_literal(self, value: Attribute) -> "TypeLiteralPosition":
        return TypeLiteralPosition(value=value, parent=None)


@dataclass(frozen=True, kw_only=True)
class UsersPosition(Position):
    """Represents users of a value"""

    use_representative: bool

    def get_for_each(self, for_each_id: int) -> "ForEachPosition":
        return ForEachPosition(parent=self, id=for_each_id)


@dataclass(frozen=True, kw_only=True)
class ForEachPosition(Position):
    """Represents an iterative choice of an operation from a set of users."""

    id: int

    def get_passthrough_op(self) -> "OperationPosition":
        return OperationPosition(depth=self.get_operation_depth() + 1, parent=self)


@dataclass(frozen=True, kw_only=True)
class ResultGroupPosition(Position):
    """Represents a group of results"""

    group_number: int | None
    is_variadic: bool

    def get_type(self) -> "TypePosition":
        return TypePosition(parent=self)


@dataclass(frozen=True, kw_only=True)
class AttributeLiteralPosition(Position):
    """Represents a literal attribute value"""

    value: Attribute


@dataclass(frozen=True, kw_only=True)
class TypeLiteralPosition(Position):
    """Represents a literal type value"""

    value: Attribute  # Can be a single type or array of types

    @staticmethod
    def get_type_literal(value: Attribute) -> "TypeLiteralPosition":
        return TypeLiteralPosition(value=value, parent=None)


@dataclass(frozen=True, kw_only=True)
class ConstraintPosition(Position):
    """Represents a result from a constraint"""

    constraint: "ConstraintQuestion"
    result_index: int

    @staticmethod
    def get_constraint(
        constraint_question: "ConstraintQuestion", result_index: int
    ) -> "ConstraintPosition":
        return ConstraintPosition(
            constraint=constraint_question, result_index=result_index, parent=None
        )


POSITION_COSTS = {
    OperationPosition: 1,
    OperandPosition: 2,
    OperandGroupPosition: 3,
    AttributePosition: 4,
    ConstraintPosition: 5,
    ResultPosition: 6,
    ResultGroupPosition: 7,
    TypePosition: 8,
    AttributeLiteralPosition: 9,
    TypeLiteralPosition: 10,
    UsersPosition: 11,
    ForEachPosition: 12,
}
"""
Different position types are ranked by priority.
A lower cost means a higher priority.
This is used to decide which position to branch on first when evaluating predicates.
"""

# =============================================================================
# Predicate System - Questions and Answers
# =============================================================================


@dataclass(frozen=True)
class Question:
    """Represents a question/check to perform"""

    pass


@dataclass(frozen=True)
class Answer:
    """Represents an expected answer to a question"""

    pass


@dataclass()
class Predicate:
    """Base predicate class"""

    q: Question
    a: Answer

    @staticmethod
    def get_is_not_null() -> "Predicate":
        return Predicate(IsNotNullQuestion(), TrueAnswer())

    @staticmethod
    def get_operation_name(name: str) -> "Predicate":
        return Predicate(OperationNameQuestion(), StringAnswer(value=name))

    @staticmethod
    def get_operand_count(count: int) -> "Predicate":
        return Predicate(OperandCountQuestion(), UnsignedAnswer(value=count))

    @staticmethod
    def get_result_count(count: int) -> "Predicate":
        return Predicate(ResultCountQuestion(), UnsignedAnswer(value=count))

    @staticmethod
    def get_equal_to(other_position: Position) -> "Predicate":
        return Predicate(EqualToQuestion(other_position=other_position), TrueAnswer())

    @staticmethod
    def get_operand_count_at_least(count: int) -> "Predicate":
        """Get predicate for minimum operand count (variadic case)"""
        return Predicate(OperandCountAtLeastQuestion(), UnsignedAnswer(value=count))

    @staticmethod
    def get_result_count_at_least(count: int) -> "Predicate":
        """Get predicate for minimum result count (variadic case)"""
        return Predicate(ResultCountAtLeastQuestion(), UnsignedAnswer(value=count))

    @staticmethod
    def get_attribute_constraint(attr_value: Attribute) -> "Predicate":
        """Get predicate for attribute value constraint"""
        return Predicate(
            AttributeConstraintQuestion(), AttributeAnswer(value=attr_value)
        )

    @staticmethod
    def get_type_constraint(
        type_value: TypeAttribute | ArrayAttr[TypeAttribute],
    ) -> "Predicate":
        """Get predicate for type value constraint"""
        return Predicate(TypeConstraintQuestion(), TypeAnswer(value=type_value))

    @staticmethod
    def get_constraint(
        name: str,
        arg_positions: tuple[Position, ...],
        result_types: tuple[pdl.AnyPDLType, ...],
        is_negated: bool = False,
    ) -> "Predicate":
        """Get predicate for a native constraint"""
        question = ConstraintQuestion(
            name=name,
            arg_positions=tuple(arg_positions),
            result_types=tuple(result_types),
            is_negated=is_negated,
        )
        return Predicate(question, TrueAnswer())


# Question Types
@dataclass(frozen=True)
class IsNotNullQuestion(Question):
    pass


@dataclass(frozen=True)
class OperationNameQuestion(Question):
    pass


@dataclass(frozen=True)
class OperandCountQuestion(Question):
    pass


@dataclass(frozen=True)
class ResultCountQuestion(Question):
    pass


@dataclass(frozen=True)
class EqualToQuestion(Question):
    other_position: Position


@dataclass(frozen=True)
class OperandCountAtLeastQuestion(Question):
    pass


@dataclass(frozen=True)
class ResultCountAtLeastQuestion(Question):
    pass


@dataclass(frozen=True)
class AttributeConstraintQuestion(Question):
    pass


@dataclass(frozen=True)
class TypeConstraintQuestion(Question):
    pass


@dataclass(frozen=True)
class ConstraintQuestion(Question):
    """Represents a native constraint check"""

    name: str
    arg_positions: tuple[Position, ...]
    result_types: tuple[pdl.AnyPDLType, ...]
    is_negated: bool


QUESTION_COSTS = {
    IsNotNullQuestion: 1,
    OperationNameQuestion: 2,
    OperandCountAtLeastQuestion: 3,
    OperandCountQuestion: 4,
    ResultCountAtLeastQuestion: 5,
    ResultCountQuestion: 6,
    EqualToQuestion: 7,
    AttributeConstraintQuestion: 8,
    TypeConstraintQuestion: 9,
    ConstraintQuestion: 10,
}
"""
Different question types are ranked by priority.
A lower cost means a higher priority.
This is used to decide which question to branch on first when evaluating predicates.
"""


def get_position_cost(position: Position) -> int:
    """Get cost for a position type, with fallback for unknown types."""
    assert (t := type(position)) in POSITION_COSTS
    return POSITION_COSTS[t]


def get_question_cost(question: Question) -> int:
    """Get cost for a question type, with fallback for unknown types."""
    assert (t := type(question)) in QUESTION_COSTS
    return QUESTION_COSTS[t]


# Answer Types
@dataclass(frozen=True)
class TrueAnswer(Answer):
    pass


@dataclass(frozen=True)
class FalseAnswer(Answer):
    pass


@dataclass(frozen=True)
class UnsignedAnswer(Answer):
    value: int


@dataclass(frozen=True)
class StringAnswer(Answer):
    value: str


@dataclass(frozen=True)
class AttributeAnswer(Answer):
    value: Attribute


@dataclass(frozen=True)
class TypeAnswer(Answer):
    value: TypeAttribute | ArrayAttr[TypeAttribute]


# =============================================================================
# Positional Predicates
# =============================================================================


@dataclass
class PositionalPredicate(Predicate):
    """A predicate applied to a specific position"""

    position: Position



from abc import ABC


class DialectInterface(ABC):
    """
    A base class for dialects' interfaces.
    They usually define functionality which is dialect specific to some transformation.

    For example DialectInlinerInterface defines which dialect operations can be inlined and how.
    Dialects will implement this interface and the inlining transformation will query them through the base interface.

    The design logic tries to follow MLIR's dialect interfaces closely
    https://mlir.llvm.org/docs/Interfaces/#dialect-interfaces
    """



"""
The pdl-to-pdl-interp conversion works by translating declarative [pdl patterns][xdsl.dialects.pdl] into a set of
[predicates][xdsl.transforms.convert_pdl_to_pdl_interp.predicate] that are then converted to imperative code in
the [pdl_interp dialect][xdsl.dialects.pdl_interp].
"""



from __future__ import annotations

import re
from collections.abc import Iterator
from dataclasses import dataclass
from enum import Enum
from typing import TypeAlias

from xdsl.utils.exceptions import PassPipelineParseError
from xdsl.utils.lexer import Input, Span, Token
from xdsl.utils.mlir_lexer import StringLiteral


class SpecTokenKind(Enum):
    EOF = object()

    IDENT = object()
    L_BRACE = "{"
    R_BRACE = "}"
    EQUALS = "="
    NUMBER = object()
    SPACE = object()
    STRING_LIT = object()
    MLIR_PIPELINE = object()
    COMMA = ","


SpecToken: TypeAlias = Token[SpecTokenKind]

_lexer_rules: list[tuple[re.Pattern[str], SpecTokenKind]] = [
    # first rule is special to allow 2d-slice to be recognized as an ident
    (re.compile(r"[0-9]+[A-Za-z_-]+[A-Za-z0-9_-]*"), SpecTokenKind.IDENT),
    (re.compile(r"[-+]?[0-9]+(\.[0-9]*([eE][-+]?[0-9]+)?)?"), SpecTokenKind.NUMBER),
    (re.compile(r"[A-Za-z0-9_-]+"), SpecTokenKind.IDENT),
    (re.compile(r'"(\\[nfvtr"\\]|[^\n\f\v\r"\\])*"'), SpecTokenKind.STRING_LIT),
    (re.compile(r'\[(\\[nfvtr"\\]|[^\n\f\v\r\]\\])*\]'), SpecTokenKind.MLIR_PIPELINE),
    (re.compile(r"\{"), SpecTokenKind.L_BRACE),
    (re.compile(r"}"), SpecTokenKind.R_BRACE),
    (re.compile(r"="), SpecTokenKind.EQUALS),
    (re.compile(r"\s+"), SpecTokenKind.SPACE),
    (re.compile(r","), SpecTokenKind.COMMA),
]
"""
This is a list of lexer rules that should be tried in this specific order to get the
next token.
"""


class PipelineLexer:
    """
    This tokenizes a pass declaration string:
    pipeline          ::= pipeline-element (`,` pipeline-element)*
    pipeline-element  ::= MLIR_PIPELINE
                        | pass-name options?
    options           ::= `{` options-element ( ` ` options-element)* `}`
    options-element   ::= key (`=` value (`,` value)* )?

    key       ::= IDENT
    pass-name ::= IDENT
    value     ::= NUMBER | BOOL | IDENT | STRING_LITERAL
    """

    _stream: Iterator[SpecToken]
    _peeked: SpecToken | None

    def __init__(self, input_str: str):
        self._stream = PipelineLexer._generator(input_str)
        self._peeked = None

    @staticmethod
    def _generator(input_str: str) -> Iterator[SpecToken]:
        input = Input(input_str, "pass-pipeline")
        pos = 0
        end = len(input_str)

        if len(input_str) == 0:
            yield SpecToken(SpecTokenKind.EOF, Span(pos, pos + 1, input))
            return

        while True:
            token: SpecToken | None = None
            for pattern, kind in _lexer_rules:
                if (match := pattern.match(input_str, pos)) is not None:
                    token = SpecToken(kind, Span(match.start(), match.end(), input))
                    pos = match.end()
                    break
            if token is None:
                raise PassPipelineParseError(
                    SpecToken(SpecTokenKind.IDENT, Span(pos, pos + 1, input)),
                    "Unknown token",
                )
            yield token
            if pos >= end:
                yield SpecToken(SpecTokenKind.EOF, Span(pos, pos + 1, input))
                return

    def lex(self) -> SpecToken:
        token = self.peek()
        self._peeked = None
        return token

    def peek(self) -> SpecToken:
        if self._peeked is None:
            self._peeked = next(self._stream)
        return self._peeked


PassArgElementType = str | int | bool | float
PassArgListType = tuple[PassArgElementType, ...]


def _pass_arg_element_type_str(arg: PassArgElementType) -> str:
    match arg:
        case bool():
            return str(arg).lower()
        case str():
            return f'"{arg}"'
        case int():
            return str(arg)
        case float():
            return str(arg)


def _pass_arg_list_type_str(name: str, arg: PassArgListType) -> str:
    if arg:
        return f"{name}={','.join(_pass_arg_element_type_str(val) for val in arg)}"
    else:
        return name


@dataclass(eq=True, frozen=True)
class PipelinePassSpec:
    """
    A pass name and its arguments.
    """

    name: str
    args: dict[str, PassArgListType]

    def normalize_arg_names(self) -> PipelinePassSpec:
        """
        This normalized all arg names by replacing `-` with `_`
        """
        new_args: dict[str, PassArgListType] = dict()
        for k, v in self.args.items():
            new_args[k.replace("-", "_")] = v
        return PipelinePassSpec(name=self.name, args=new_args)

    def __str__(self) -> str:
        """
        This function returns a string containing the PipelineSpec name, its arguments
        and respective values for use on the commandline.
        """
        query = f"{self.name}"
        arguments_pipeline = " ".join(
            _pass_arg_list_type_str(arg_name, arg_val)
            for arg_name, arg_val in self.args.items()
        )
        query += f"{{{arguments_pipeline}}}" if self.args else ""

        return query


def parse_pipeline(
    pipeline_spec: str,
) -> Iterator[PipelinePassSpec]:
    """
    This takes a pipeline string and gives a representation of
    the specification.

    Each pass is represented by a tuple of:
     - name: the name of the pass as string
     - args: a dictionary, where each value is zero or more
            of (str | bool | float | int)
    """
    lexer = PipelineLexer(pipeline_spec)

    while True:
        # get the pass name
        name = lexer.lex()
        if name.kind is SpecTokenKind.EOF:
            return
        if name.kind is not SpecTokenKind.IDENT:
            raise PassPipelineParseError(name, "Expected pass name here")

        # valid next tokens are EOF, COMMA or `{`
        match lexer.lex():
            case Token(kind=SpecTokenKind.EOF):
                # EOF means we have nothing else left to parse, we are done
                yield PipelinePassSpec(name.span.text, dict())
                return
            case Token(kind=SpecTokenKind.COMMA):
                # comma means we are done parsing this pass, move on to next pass
                yield PipelinePassSpec(name.span.text, dict())
                continue
            case Token(kind=SpecTokenKind.L_BRACE):
                # `{` indicates start of args dict, so we parse that next
                yield PipelinePassSpec(name.span.text, _parse_pass_args(lexer))
            case Token(SpecTokenKind.MLIR_PIPELINE, span):
                if name.span.text != "mlir-opt":
                    raise PassPipelineParseError(
                        name,
                        "Expected `mlir-opt` to mark an MLIR pipeline here",
                    )
                yield PipelinePassSpec(
                    "mlir-opt",
                    {
                        "arguments": (
                            "--mlir-print-op-generic",
                            "--allow-unregistered-dialect",
                            "-p",
                            f"builtin.module({span.text[1:-1]})",
                        )
                    },
                )
            case invalid:
                # every other token is invalid
                raise PassPipelineParseError(
                    invalid, "Expected a comma or pass arguments here"
                )

        # check for comma or EOF
        match lexer.lex():
            case Token(kind=SpecTokenKind.EOF):
                # EOF means we are finished parsing
                return
            case Token(kind=SpecTokenKind.COMMA):
                # comma means we move on to parse the next pass spec
                continue
            case invalid:
                # every other token is invalid
                raise PassPipelineParseError(
                    invalid, "Expected a comma after pass argument dict here"
                )


def _parse_pass_args(lexer: PipelineLexer) -> dict[str, PassArgListType]:
    """
    This parses pass arguments. They are a dictionary structure
    with whitespace separated, multi-value elements:

    options           ::= `{` options-element ( ` ` options-element)* `}`
    options-element   ::= key (`=` value (`,` value)* )?

    This function assumes that the leading `{` has already been consumed.
    """
    args: dict[str, PassArgListType] = dict()

    while True:
        # get the name of the argument (or a `}` in case of zero-length dicts)
        name = lexer.lex()

        # allow for zero-length arg dicts
        if name.kind is SpecTokenKind.R_BRACE:
            return args

        # check that it is a valid identifier
        if name.kind is not SpecTokenKind.IDENT:
            raise PassPipelineParseError(name, "Expected argument name here")

        # next token should be either a space, `}` or `=`
        match lexer.lex():
            case Token(kind=SpecTokenKind.SPACE):
                # space means zero-length argument, store empty list
                args[name.span.text] = ()
                # then continue parsing args list
                continue
            case Token(kind=SpecTokenKind.R_BRACE):
                # `}` means zero-length argument with no further arg
                args[name.span.text] = ()
                # stop parsing args
                return args
            case Token(kind=SpecTokenKind.EQUALS):
                # equals means we have an arg value given, parse it
                args[name.span.text] = _parse_arg_value(lexer)
            case invalid:
                # every other token is invalid
                raise PassPipelineParseError(
                    invalid, "Expected equals, space or end of arguments here"
                )

        # next token must be either space or `}`
        match lexer.lex():
            case Token(kind=SpecTokenKind.SPACE):
                # space means we get another argument
                continue
            case Token(kind=SpecTokenKind.R_BRACE):
                # `}` signifies end of args
                return args
            case invalid:
                # every other token is a syntax error
                raise PassPipelineParseError(
                    invalid,
                    "Malformed pass arguments, expected either a space or `}` here",
                )


def _parse_arg_value(lexer: PipelineLexer) -> PassArgListType:
    """
    Parse an argument value of the form: value (`,` value)*
    """
    elms = [_parse_arg_value_element(lexer)]
    while lexer.peek().kind is SpecTokenKind.COMMA:
        lexer.lex()
        elms.append(_parse_arg_value_element(lexer))
    return tuple(elms)


def _parse_arg_value_element(lexer: PipelineLexer) -> PassArgElementType:
    """
    Parse a singular value element
    """
    # valid value elements are quoted strings, numbers, true|false, and "ident" type
    # strings
    match lexer.lex():
        case Token(kind=SpecTokenKind.STRING_LIT, span=span):
            # string literals are converted to unescaped strings
            str_token = StringLiteral.from_span(span)
            assert str_token is not None
            return str_token.string_contents
        case Token(kind=SpecTokenKind.NUMBER, span=span):
            # NUMBER is both float and int
            # if the token contains a `.` it's a float
            if "." in span.text:
                return float(span.text)
            # otherwise an int
            return int(span.text)
        case Token(kind=SpecTokenKind.IDENT, span=span):
            # identifiers are either true|false or treated as a string
            if span.text == "true":
                return True
            elif span.text == "false":
                return False
            return span.text
        case token:
            # every other token type is invalid as a value
            raise PassPipelineParseError(
                token,
                "Unknown argument value, wrap argument in quotes to pass arbitrary "
                "string values",
            )



from abc import ABC, abstractmethod

from xdsl.dialect_interfaces import DialectInterface
from xdsl.ir import Attribute, Operation


class ConstantMaterializationInterface(DialectInterface, ABC):
    """
    An interface for dialects that support constant materialization.

    A dialect that implements this interface should provide the `materialize_constant` method,
    which creates a constant operation of the dialect given a value and a type.

    This is useful for transformations that need to create constants in a dialect-specific way.
    """

    @abstractmethod
    def materialize_constant(
        self, value: Attribute, type: Attribute
    ) -> Operation | None:
        """
        Materializes a constant operation in the dialect.

        Args:
            value (Attribute): The attribute representing the constant value.
            type (Attribute): The type of the constant.

        Returns:
            Operation: The created constant operation.
        """
        raise NotImplementedError("Dialect does not implement materialize_constant")



# RUN: python %s | filecheck %s

from ctypes import c_size_t

from xdsl.dialects import bigint, builtin
from xdsl.frontend.pyast.context import PyASTContext
from xdsl.frontend.pyast.utils.exceptions import CodeGenerationException

ctx = PyASTContext(post_transforms=[])
ctx.register_type(bool, builtin.i1)
ctx.register_type(int, bigint.bigint)
ctx.register_type(c_size_t, builtin.IndexType())
ctx.register_type(float, builtin.f64)


# CHECK:      %{{.*}} = scf.if %{{.*}} -> (!bigint.bigint) {
# CHECK-NEXT:   %{{.*}} = symref.fetch @x : !bigint.bigint
# CHECK-NEXT:   scf.yield %{{.*}} : !bigint.bigint
# CHECK-NEXT: } else {
# CHECK-NEXT:   %{{.*}} = symref.fetch @y : !bigint.bigint
# CHECK-NEXT:   scf.yield %{{.*}} : !bigint.bigint
# CHECK-NEXT: }
@ctx.parse_program
def test_if_expr(cond: bool, x: int, y: int) -> int:
    return x if cond else y


print(test_if_expr.module)


# CHECK:      scf.if %{{.*}} {
# CHECK-NEXT: }
@ctx.parse_program
def test_if_I(cond: bool):
    if cond:
        pass
    else:
        pass
    return


print(test_if_I.module)


# CHECK:      %{{.*}} = symref.fetch @a : i1
# CHECK-NEXT: scf.if %{{.*}} {
# CHECK-NEXT: } else {
# CHECK-NEXT:   %{{.*}} = symref.fetch @b : i1
# CHECK-NEXT:   scf.if %{{.*}} {
# CHECK-NEXT:   } else {
# CHECK-NEXT:     %{{.*}} = symref.fetch @c : i1
# CHECK-NEXT:     scf.if %{{.*}} {
# CHECK-NEXT:     }
# CHECK-NEXT:   }
# CHECK-NEXT: }
@ctx.parse_program
def test_if_II(a: bool, b: bool, c: bool):
    if a:
        pass
    elif b:
        pass
    elif c:
        pass
    return


print(test_if_II.module)


# CHECK:      %{{.*}} = symref.fetch @cond : i1
# CHECK-NEXT: scf.if %{{.*}} {
# CHECK-NEXT: }
@ctx.parse_program
def test_if_III(cond: bool):
    if cond:
        pass
    return


print(test_if_III.module)


# CHECK: Expected the same types for if expression, but got !bigint.bigint and f64.
@ctx.parse_program
def test_type_mismatch_in_if_expr(cond: bool, x: int, y: float) -> int:
    return x if cond else y  # pyright: ignore[reportReturnType]


try:
    test_type_mismatch_in_if_expr.module
except CodeGenerationException as e:
    print(e.msg)
